import dgl
import torch
import typing as _typing
from . import (
    _abstract_views,
    _canonical_edge_type,
    _general_static_graph
)


class _DGLGraphHolder:
    def __init__(self, dgl_graph: dgl.DGLGraph):
        if not isinstance(dgl_graph, dgl.DGLGraph):
            raise TypeError
        self.__graph: dgl.DGLGraph = dgl_graph

    @property
    def graph(self) -> dgl.DGLGraph:
        return self.__graph

    @graph.setter
    def graph(self, dgl_graph: dgl.DGLGraph):
        if not isinstance(dgl_graph, dgl.DGLGraph):
            raise TypeError
        else:
            self.__graph = dgl_graph


class _SpecificTypedNodeDataView(_abstract_views.SpecificTypedNodeDataView):
    def __init__(
            self, dgl_graph_holder: _DGLGraphHolder,
            node_type: _typing.Optional[str] = ...
    ):
        if not isinstance(dgl_graph_holder, _DGLGraphHolder):
            raise TypeError
        if not (node_type in (Ellipsis, None) or isinstance(node_type, str)):
            raise TypeError
        elif isinstance(node_type, str) and ' ' in node_type:
            raise ValueError("Illegal node type")
        self.__dgl_graph_holder: _DGLGraphHolder = dgl_graph_holder
        self.__optional_node_type: _typing.Optional[str] = (
            node_type if isinstance(node_type, str) else None
        )

    def __getitem__(self, data_key: str) -> torch.Tensor:
        if not isinstance(data_key, str):
            raise TypeError
        elif ' ' in data_key:
            raise ValueError("Illegal data key")
        if isinstance(self.__optional_node_type, str):
            node_type: str = self.__optional_node_type
        else:
            if len(self.__dgl_graph_holder.graph.ntypes) == 0:
                raise ValueError("the graph is empty")
            elif len(self.__dgl_graph_holder.graph.ntypes) > 1:
                raise ValueError(
                    "Unable to automatically determine node type, "
                    "the graph consists of heterogeneous node types"
                )
            else:
                node_type: str = self.__dgl_graph_holder.graph.ntypes[0]
        if data_key in self.__dgl_graph_holder.graph.nodes[node_type].data:
            return self.__dgl_graph_holder.graph.nodes[node_type].data[data_key]
        else:
            raise KeyError  # todo: Complete message

    def __setitem__(self, data_key: str, value: torch.Tensor):
        if not isinstance(data_key, str):
            raise TypeError
        elif ' ' in data_key:
            raise ValueError("Illegal data key")
        if not isinstance(value, torch.Tensor):
            raise TypeError
        elif value.dim() == 0:
            raise ValueError
        if isinstance(self.__optional_node_type, str):
            node_type: str = self.__optional_node_type
        else:
            if len(self.__dgl_graph_holder.graph.ntypes) == 0:
                raise ValueError("the graph is empty")
            elif len(self.__dgl_graph_holder.graph.ntypes) > 1:
                raise ValueError(
                    "Unable to automatically determine node type, "
                    "the graph consists of heterogeneous node types"
                )
            else:
                node_type: str = self.__dgl_graph_holder.graph.ntypes[0]
        if value.size(0) != self.__dgl_graph_holder.graph.num_nodes(node_type):
            raise ValueError  # todo: Complete error message
        else:
            # todo: 现在这个方法没有处理node_type不存在的情况
            self.__dgl_graph_holder.graph.nodes[node_type].data[data_key] = value

    def __delitem__(self, data_key: str) -> None:
        if not isinstance(data_key, str):
            raise TypeError
        elif ' ' in data_key:
            raise ValueError("Illegal data key")
        if isinstance(self.__optional_node_type, str):
            node_type: str = self.__optional_node_type
        else:
            if len(self.__dgl_graph_holder.graph.ntypes) == 0:
                raise ValueError("the graph is empty")
            elif len(self.__dgl_graph_holder.graph.ntypes) > 1:
                raise ValueError(
                    "Unable to automatically determine node type, "
                    "the graph consists of heterogeneous node types"
                )
            else:
                node_type: str = self.__dgl_graph_holder.graph.ntypes[0]
        if data_key in self.__dgl_graph_holder.graph.nodes[node_type].data:
            try:
                del self.__dgl_graph_holder.graph.nodes[node_type].data[data_key]
            except KeyError:
                pass  # todo: Use logger to warn

    def __len__(self) -> int:
        if isinstance(self.__optional_node_type, str):
            node_type: str = self.__optional_node_type
        else:
            if len(self.__dgl_graph_holder.graph.ntypes) == 0:
                raise ValueError("the graph is empty")
            elif len(self.__dgl_graph_holder.graph.ntypes) > 1:
                raise ValueError(
                    "Unable to automatically determine node type, "
                    "the graph consists of heterogeneous node types"
                )
            else:
                node_type: str = self.__dgl_graph_holder.graph.ntypes[0]
        return len(self.__dgl_graph_holder.graph.nodes[node_type].data)

    def __iter__(self) -> _typing.Iterator[str]:
        if isinstance(self.__optional_node_type, str):
            node_type: str = self.__optional_node_type
        else:
            if len(self.__dgl_graph_holder.graph.ntypes) == 0:
                raise ValueError("the graph is empty")
            elif len(self.__dgl_graph_holder.graph.ntypes) > 1:
                raise ValueError(
                    "Unable to automatically determine node type, "
                    "the graph consists of heterogeneous node types"
                )
            else:
                node_type: str = self.__dgl_graph_holder.graph.ntypes[0]
        return iter(self.__dgl_graph_holder.graph.nodes[node_type].data)


class _SpecificTypedNodeView(_abstract_views.SpecificTypedNodeView):
    def __init__(
            self, dgl_graph_holder: _DGLGraphHolder,
            node_type: _typing.Optional[str] = ...
    ):
        if not isinstance(dgl_graph_holder, _DGLGraphHolder):
            raise TypeError
        if not (node_type in (Ellipsis, None) or isinstance(node_type, str)):
            raise TypeError
        elif isinstance(node_type, str) and ' ' in node_type:
            raise ValueError("Illegal node type")
        self.__dgl_graph_holder: _DGLGraphHolder = dgl_graph_holder
        self.__optional_node_type: _typing.Optional[str] = (
            node_type if isinstance(node_type, str) else None
        )

    @property
    def data(self) -> _SpecificTypedNodeDataView:
        return _SpecificTypedNodeDataView(
            self.__dgl_graph_holder, self.__optional_node_type
        )

    @data.setter
    def data(self, nodes_data: _typing.Mapping[str, torch.Tensor]):
        raise NotImplementedError  # todo: Currently, DGL not support this operation


class _HeterogeneousNodeView(_abstract_views.HeterogeneousNodeView):
    def __init__(self, dgl_graph_holder: _DGLGraphHolder):
        if not isinstance(dgl_graph_holder, _DGLGraphHolder):
            raise TypeError
        self.__dgl_graph_holder: _DGLGraphHolder = dgl_graph_holder

    @property
    def data(self) -> _SpecificTypedNodeDataView:
        return _SpecificTypedNodeDataView(self.__dgl_graph_holder, ...)

    @data.setter
    def data(self, nodes_data: _typing.Mapping[str, torch.Tensor]):
        if not isinstance(nodes_data, _typing.Mapping):
            raise TypeError
        _SpecificTypedNodeView(self.__dgl_graph_holder, ...).data = nodes_data

    def __getitem__(self, node_type: _typing.Optional[str]) -> _SpecificTypedNodeView:
        if not (node_type in (Ellipsis, None) or isinstance(node_type, str)):
            raise TypeError
        elif isinstance(node_type, str) and ' ' in node_type:
            raise ValueError("Illegal edge type")
        return _SpecificTypedNodeView(self.__dgl_graph_holder, node_type)

    def __setitem__(
            self, node_type: _typing.Optional[str],
            nodes_data: _typing.Mapping[str, torch.Tensor]
    ):
        if not (node_type in (Ellipsis, None) or isinstance(node_type, str)):
            raise TypeError
        elif isinstance(node_type, str) and ' ' in node_type:
            raise ValueError("Illegal edge type")
        if not isinstance(nodes_data, _typing.Mapping):
            raise TypeError
        _SpecificTypedNodeView(
            self.__dgl_graph_holder, node_type if isinstance(node_type, str) else None
        ).data = nodes_data

    def __delitem__(self, node_t: _typing.Optional[str]):
        raise NotImplementedError  # todo: Currently, DGL not support this operation

    def __iter__(self) -> _typing.Iterator[str]:
        return iter(self.__dgl_graph_holder.graph.ntypes)

    @property
    def is_homogeneous(self) -> bool:
        return len(self.__dgl_graph_holder.graph.ntypes) <= 1


class _HomogeneousEdgesDataView(_abstract_views.HomogeneousEdgesDataView):
    def __init__(
            self, dgl_graph_holder: _DGLGraphHolder,
            edge_type: _typing.Union[
                None, str, _typing.Tuple[str, str, str],
                _canonical_edge_type.CanonicalEdgeType
            ] = ...
    ):
        if not isinstance(dgl_graph_holder, _DGLGraphHolder):
            raise TypeError
        self.__dgl_graph_holder: _DGLGraphHolder = dgl_graph_holder
        if edge_type in (Ellipsis, None):
            self.__optional_edge_type: _typing.Union[None, str, _typing.Tuple[str, str, str]] = None
        elif isinstance(edge_type, str):
            if ' ' in edge_type:
                raise ValueError("Illegal edge type")
            self.__optional_edge_type: _typing.Union[None, str, _typing.Tuple[str, str, str]] = edge_type
        elif isinstance(edge_type, _typing.Sequence) and not isinstance(edge_type, str):
            if not (
                    len(edge_type) == 3 and
                    isinstance(edge_type[0], str) and ' ' not in edge_type[0] and
                    isinstance(edge_type[1], str) and ' ' not in edge_type[1] and
                    isinstance(edge_type[2], str) and ' ' not in edge_type[2]
            ):
                raise ValueError("Illegal edge type")
            self.__optional_edge_type: _typing.Union[None, str, _typing.Tuple[str, str, str]] = tuple(edge_type)
        elif isinstance(edge_type, _canonical_edge_type.CanonicalEdgeType):
            self.__optional_edge_type: _typing.Union[None, str, _typing.Tuple[str, str, str]] = (
                edge_type.source_node_type, edge_type.relation_type, edge_type.target_node_type
            )
        else:
            raise TypeError

    def __get_canonical_edge_type(self) -> _typing.Tuple[str, str, str]:
        if self.__optional_edge_type in (Ellipsis, None):
            if len(self.__dgl_graph_holder.graph.canonical_etypes) == 0:
                raise ValueError("The graph is empty")
            elif len(self.__dgl_graph_holder.graph.canonical_etypes) > 1:
                raise ValueError(
                    "Unable to automatically determine edge type, "
                    "the graph consists of heterogeneous edge types."
                )
            else:
                return self.__dgl_graph_holder.graph.canonical_etypes[0]
        elif isinstance(self.__optional_edge_type, str):
            try:
                canonical_edge_type = self.__dgl_graph_holder.graph.to_canonical_etype(
                    self.__optional_edge_type
                )
            except dgl.DGLError as e:
                raise e
            else:
                return canonical_edge_type
        else:
            return self.__optional_edge_type

    def __getitem__(self, data_key: str) -> torch.Tensor:
        if not isinstance(data_key, str):
            raise TypeError
        elif ' ' in data_key:
            raise ValueError("Illegal data key")
        edge_type: _typing.Tuple[str, str, str] = self.__get_canonical_edge_type()

        found = False
        for et in self.__dgl_graph_holder.graph.canonical_etypes:
            if all([a == b for a, b in zip(et, edge_type)]):
                found = True
                break
        if not found:
            raise ValueError("edge type not exist")

        if data_key in self.__dgl_graph_holder.graph.edges[edge_type].data:
            return self.__dgl_graph_holder.graph.edges[edge_type].data[data_key]
        else:
            raise KeyError  # todo: Complete error message

    def __setitem__(self, data_key: str, value: torch.Tensor):
        if not isinstance(data_key, str):
            raise TypeError
        elif ' ' in data_key:
            raise ValueError("Illegal data key")
        if not isinstance(value, torch.Tensor):
            raise TypeError
        if value.dim() == 0:
            raise ValueError
        edge_type: _typing.Tuple[str, str, str] = self.__get_canonical_edge_type()

        found = False
        for et in self.__dgl_graph_holder.graph.canonical_etypes:
            if all([a == b for a, b in zip(et, edge_type)]):
                found = True
                break
        if not found:
            raise ValueError("edge type not exist")

        self.__dgl_graph_holder.graph.edges[edge_type].data[data_key] = value

    def __delitem__(self, data_key: str):
        if not isinstance(data_key, str):
            raise TypeError
        elif ' ' in data_key:
            raise ValueError("Illegal data key")
        edge_type: _typing.Tuple[str, str, str] = self.__get_canonical_edge_type()

        found = False
        for et in self.__dgl_graph_holder.graph.canonical_etypes:
            if all([a == b for a, b in zip(et, edge_type)]):
                found = True
                break
        if not found:
            raise ValueError("edge type not exist")

        if data_key in self.__dgl_graph_holder.graph.edges[edge_type].data:
            del self.__dgl_graph_holder.graph.edges[edge_type].data[data_key]
        else:
            raise KeyError  # todo: Complete error message

    def __len__(self) -> int:
        edge_type: _typing.Tuple[str, str, str] = self.__get_canonical_edge_type()

        found = False
        for et in self.__dgl_graph_holder.graph.canonical_etypes:
            if all([a == b for a, b in zip(et, edge_type)]):
                found = True
                break
        if not found:
            raise ValueError("edge type not exist")

        return len(self.__dgl_graph_holder.graph.edges[edge_type].data)

    def __iter__(self) -> _typing.Iterator[str]:
        edge_type: _typing.Tuple[str, str, str] = self.__get_canonical_edge_type()

        found = False
        for et in self.__dgl_graph_holder.graph.canonical_etypes:
            if all([a == b for a, b in zip(et, edge_type)]):
                found = True
                break
        if not found:
            raise ValueError("edge type not exist")

        return iter(self.__dgl_graph_holder.graph.edges[edge_type].data)


class _HomogeneousEdgesView(_abstract_views.HomogeneousEdgesView):
    def __init__(
            self, dgl_graph_holder: _DGLGraphHolder,
            edge_type: _typing.Union[
                None, str, _typing.Tuple[str, str, str],
                _canonical_edge_type.CanonicalEdgeType
            ] = ...
    ):
        if not isinstance(dgl_graph_holder, _DGLGraphHolder):
            raise TypeError
        self.__dgl_graph_holder: _DGLGraphHolder = dgl_graph_holder
        if edge_type in (Ellipsis, None):
            self.__optional_edge_type: _typing.Union[None, str, _typing.Tuple[str, str, str]] = None
        elif isinstance(edge_type, str):
            if ' ' in edge_type:
                raise ValueError("Illegal edge type")
            self.__optional_edge_type: _typing.Union[None, str, _typing.Tuple[str, str, str]] = edge_type
        elif isinstance(edge_type, _typing.Sequence) and not isinstance(edge_type, str):
            if not (
                    len(edge_type) == 3 and
                    isinstance(edge_type[0], str) and ' ' not in edge_type[0] and
                    isinstance(edge_type[1], str) and ' ' not in edge_type[1] and
                    isinstance(edge_type[2], str) and ' ' not in edge_type[2]
            ):
                raise ValueError("Illegal edge type")
            self.__optional_edge_type: _typing.Union[None, str, _typing.Tuple[str, str, str]] = tuple(edge_type)
        elif isinstance(edge_type, _canonical_edge_type.CanonicalEdgeType):
            self.__optional_edge_type: _typing.Union[None, str, _typing.Tuple[str, str, str]] = (
                edge_type.source_node_type, edge_type.relation_type, edge_type.target_node_type
            )
        else:
            raise TypeError

    def __get_canonical_edge_type(self) -> _typing.Tuple[str, str, str]:
        if self.__optional_edge_type in (Ellipsis, None):
            if len(self.__dgl_graph_holder.graph.canonical_etypes) == 0:
                raise ValueError("The graph is empty")
            elif len(self.__dgl_graph_holder.graph.canonical_etypes) > 1:
                raise ValueError(
                    "Unable to automatically determine edge type, "
                    "the graph consists of heterogeneous edge types."
                )
            else:
                return self.__dgl_graph_holder.graph.canonical_etypes[0]
        elif isinstance(self.__optional_edge_type, str):
            try:
                canonical_edge_type = self.__dgl_graph_holder.graph.to_canonical_etype(
                    self.__optional_edge_type
                )
            except dgl.DGLError as e:
                raise e
            else:
                return canonical_edge_type
        else:
            return self.__optional_edge_type

    @property
    def connections(self) -> torch.Tensor:
        return torch.vstack(
            self.__dgl_graph_holder.graph.edges(etype=self.__get_canonical_edge_type())
        )

    @property
    def data(self) -> _HomogeneousEdgesDataView:
        return _HomogeneousEdgesDataView(self.__dgl_graph_holder, self.__optional_edge_type)


class _HeterogeneousEdgesView(_abstract_views.HeterogeneousEdgesView):
    def __init__(self, dgl_graph_holder: _DGLGraphHolder):
        if not isinstance(dgl_graph_holder, _DGLGraphHolder):
            raise TypeError
        self.__dgl_graph_holder: _DGLGraphHolder = dgl_graph_holder
        self.__optional_edge_type: _typing.Union[None, str, _typing.Tuple[str, str, str]] = None

    def __get_canonical_edge_type(self) -> _typing.Tuple[str, str, str]:
        if self.__optional_edge_type in (Ellipsis, None):
            if len(self.__dgl_graph_holder.graph.canonical_etypes) == 0:
                raise ValueError("The graph is empty")
            elif len(self.__dgl_graph_holder.graph.canonical_etypes) > 1:
                raise ValueError(
                    "Unable to automatically determine edge type, "
                    "the graph consists of heterogeneous edge types."
                )
            else:
                return self.__dgl_graph_holder.graph.canonical_etypes[0]
        elif isinstance(self.__optional_edge_type, str):
            try:
                canonical_edge_type = self.__dgl_graph_holder.graph.to_canonical_etype(
                    self.__optional_edge_type
                )
            except dgl.DGLError as e:
                raise e
            else:
                return canonical_edge_type
        else:
            return self.__optional_edge_type

    @property
    def connections(self) -> torch.Tensor:
        return _HomogeneousEdgesView(self.__dgl_graph_holder, ...).connections

    @property
    def data(self) -> _HomogeneousEdgesDataView:
        return _HomogeneousEdgesView(self.__dgl_graph_holder, ...).data

    @property
    def is_homogeneous(self) -> bool:
        return len(self.__dgl_graph_holder.graph.canonical_etypes) <= 1

    def set(
            self, edge_t: _typing.Union[None, str, _typing.Tuple[str, str, str]],
            connections: torch.LongTensor,
            data: _typing.Optional[_typing.Mapping[str, torch.Tensor]] = ...
    ):
        raise NotImplementedError  # todo: Complete this function or this error message

    def __getitem__(
            self,
            edge_t: _typing.Union[
                None, str, _typing.Tuple[str, str, str], _canonical_edge_type.CanonicalEdgeType
            ]
    ) -> _HomogeneousEdgesView:
        return _HomogeneousEdgesView(self.__dgl_graph_holder, edge_t)

    def __setitem__(
            self,
            edge_t: _typing.Union[
                None, str, _typing.Tuple[str, str, str], _canonical_edge_type.CanonicalEdgeType
            ],
            edges: _typing.Union[torch.LongTensor]
    ):
        raise NotImplementedError  # todo: Complete this function or this error message

    def __delitem__(
            self,
            edge_t: _typing.Union[
                None, str, _typing.Tuple[str, str, str], _canonical_edge_type.CanonicalEdgeType
            ]
    ):
        raise NotImplementedError  # todo: Complete this function or this error message

    def __len__(self) -> int:
        return len(self.__dgl_graph_holder.graph.canonical_etypes)

    def __iter__(self) -> _typing.Iterator[_canonical_edge_type.CanonicalEdgeType]:
        return iter([
            _canonical_edge_type.CanonicalEdgeType(et[0], et[1], et[2])
            for et in self.__dgl_graph_holder.graph.canonical_etypes
        ])

    def __contains__(
            self,
            edge_type: _typing.Union[
                str, _typing.Tuple[str, str, str], _canonical_edge_type.CanonicalEdgeType
            ]
    ) -> bool:
        # raise NotImplementedError
        if isinstance(edge_type, str):
            if ' ' in edge_type:
                raise ValueError("Illegal edge type")
            else:
                return edge_type in self.__dgl_graph_holder.graph.etypes
        elif isinstance(edge_type, _typing.Sequence) and not isinstance(edge_type, str):
            if not (
                    len(edge_type) == 3 and
                    isinstance(edge_type[0], str) and ' ' not in edge_type[0] and
                    isinstance(edge_type[1], str) and ' ' not in edge_type[1] and
                    isinstance(edge_type[2], str) and ' ' not in edge_type[2]
            ):
                raise ValueError("Illegal edge type")
            found = False
            for et in self.__dgl_graph_holder.graph.canonical_etypes:
                if all([a == b for a, b in zip(et, edge_type)]):
                    found = True
                    break
            return found
        elif isinstance(edge_type, _canonical_edge_type.CanonicalEdgeType):
            found = False
            for et in self.__dgl_graph_holder.graph.canonical_etypes:
                if (
                        et[0] == edge_type.source_node_type and
                        et[1] == edge_type.relation_type and
                        et[2] == edge_type.target_node_type
                ):
                    found = True
                    break
            return found
        else:
            raise TypeError


class _StaticGraphDataContainer(_typing.MutableMapping[str, torch.Tensor]):
    def __setitem__(self, data_key: str, data: torch.Tensor) -> None:
        raise NotImplementedError

    def __delitem__(self, data_key: str) -> None:
        raise NotImplementedError

    def __getitem__(self, data_key: str) -> torch.Tensor:
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError

    def __iter__(self) -> _typing.Iterator[str]:
        raise NotImplementedError


class StaticGraphDataAggregation(_StaticGraphDataContainer):
    def __init__(
            self, graph_data: _typing.Optional[_typing.Mapping[str, torch.Tensor]] = ...
    ):
        self.__data: _typing.MutableMapping[str, torch.Tensor] = (
            dict(graph_data) if isinstance(graph_data, _typing.Mapping)
            else {}
        )

    def __setitem__(self, data_key: str, data: torch.Tensor) -> None:
        self.__data[data_key] = data

    def __delitem__(self, data_key: str) -> None:
        del self.__data[data_key]

    def __getitem__(self, data_key: str) -> torch.Tensor:
        return self.__data[data_key]

    def __len__(self) -> int:
        return len(self.__data)

    def __iter__(self) -> _typing.Iterator[str]:
        return iter(self.__data)


class _StaticGraphDataView(_abstract_views.GraphDataView):
    def __init__(self, graph_data_container: _StaticGraphDataContainer):
        self.__graph_data_container: _StaticGraphDataContainer = (
            graph_data_container
        )

    def __setitem__(self, data_key: str, data: torch.Tensor) -> None:
        self.__graph_data_container[data_key] = data

    def __delitem__(self, data_key: str) -> None:
        del self.__graph_data_container[data_key]

    def __getitem__(self, data_key: str) -> torch.Tensor:
        return self.__graph_data_container[data_key]

    def __len__(self) -> int:
        return len(self.__graph_data_container)

    def __iter__(self) -> _typing.Iterator[str]:
        return iter(self.__graph_data_container)


class GeneralStaticGraphDGLImplementation(
    _general_static_graph.GeneralStaticGraph
):
    def __init__(
            self, dgl_graph: dgl.DGLGraph,
            graph_data: _typing.Optional[_typing.Mapping[str, torch.Tensor]] = ...
    ):
        if not isinstance(dgl_graph, dgl.DGLGraph) and (
                graph_data in (Ellipsis, None) or
                isinstance(graph_data, _typing.Mapping)
        ):
            raise TypeError
        self.__dgl_graph_holder: _DGLGraphHolder = _DGLGraphHolder(dgl_graph)
        self.__graph_data_container: _StaticGraphDataContainer = (
            StaticGraphDataAggregation(
                graph_data if isinstance(graph_data, _typing.Mapping) else None
            )
        )

    @property
    def nodes(self) -> _abstract_views.HeterogeneousNodeView:
        return _HeterogeneousNodeView(self.__dgl_graph_holder)

    @property
    def edges(self) -> _abstract_views.HeterogeneousEdgesView:
        return _HeterogeneousEdgesView(self.__dgl_graph_holder)

    @property
    def data(self) -> _abstract_views.GraphDataView:
        return _StaticGraphDataView(self.__graph_data_container)



from ._model_registry import MODEL_DICT, ModelUniversalRegistry, register_model
from .base import BaseAutoModel
from .topkpool import AutoTopkpool

# from .graph_sage import AutoSAGE
from .graphsage import AutoSAGE
from .graph_saint import GraphSAINTAggregationModel
from .gcn import AutoGCN
from .gat import AutoGAT
from .gin import AutoGIN

from .robust.gnnguard import AutoGNNGuard, AutoGNNGuard_attack, GCN4GNNGuard, GCN4GNNGuard_attack

__all__ = [
    "ModelUniversalRegistry",
    "register_model",
    "BaseAutoModel",
    "AutoTopkpool",
    "AutoSAGE",
    "GraphSAINTAggregationModel",
    "AutoGCN",
    "AutoGAT",
    "AutoGIN",
    "AutoGNNGuard",
    "AutoGNNGuard_attack",
    "GCN4GNNGuard",
    "GCN4GNNGuard_attack",
]



import copy
import torch
import typing
import autogl.data.graph


class _AbstractDataPreprocessor:
    def fit(self, dataset):
        raise NotImplementedError

    def transform(self, dataset, inplace: bool = True):
        raise NotImplementedError

    def fit_transform(self, dataset, inplace: bool = True):
        raise NotImplementedError


class _ComposedDataPreprocessor(_AbstractDataPreprocessor):
    def __init__(self, data_preprocessors: typing.Iterable[_AbstractDataPreprocessor]):
        self.__data_preprocessors: typing.MutableSequence[_AbstractDataPreprocessor] = []
        for preprocessor in data_preprocessors:
            if isinstance(preprocessor, _ComposedDataPreprocessor):
                self.__data_preprocessors.extend(preprocessor.__data_preprocessors)
            else:
                self.__data_preprocessors.append(preprocessor)

    def fit(self, dataset):
        for preprocessor in self.__data_preprocessors:
            dataset = preprocessor.fit(dataset)
        return dataset

    def transform(self, dataset, inplace: bool = True):
        for preprocessor in self.__data_preprocessors:
            dataset = preprocessor.transform(dataset, inplace)
        return dataset

    def fit_transform(self, dataset, inplace: bool = True):
        for preprocessor in self.__data_preprocessors:
            dataset = preprocessor.fit_transform(dataset, inplace)
        return dataset

    def __and__(self, other: _AbstractDataPreprocessor):
        return _ComposedDataPreprocessor((self, other))


class _DataPreprocessor(_AbstractDataPreprocessor):
    def __and__(self, other):
        return _ComposedDataPreprocessor((self, other))

    def _preprocess(
            self, data: typing.Union[autogl.data.graph.GeneralStaticGraph, typing.Any]
    ) -> typing.Union[autogl.data.graph.GeneralStaticGraph, typing.Any]:
        raise NotImplementedError

    def _postprocess(
            self, data: typing.Union[autogl.data.graph.GeneralStaticGraph, typing.Any]
    ) -> typing.Union[autogl.data.graph.GeneralStaticGraph, typing.Any]:
        raise NotImplementedError

    def _fit(
            self, data: typing.Union[autogl.data.graph.GeneralStaticGraph, typing.Any]
    ) -> typing.Union[autogl.data.graph.GeneralStaticGraph, typing.Any]:
        raise NotImplementedError

    def _transform(
            self, data: typing.Union[autogl.data.graph.GeneralStaticGraph, typing.Any]
    ) -> typing.Union[autogl.data.graph.GeneralStaticGraph, typing.Any]:
        raise NotImplementedError

    def fit(self, dataset):
        raise NotImplementedError

    def transform(self, dataset, inplace: bool = True):
        raise NotImplementedError

    def fit_transform(self, dataset, inplace: bool = True):
        raise NotImplementedError


class DataPreprocessor(_DataPreprocessor):
    def fit(self, dataset):
        with torch.no_grad():
            for i, data in enumerate(dataset):
                dataset[i] = self._postprocess(self._fit(self._preprocess(data)))
            return dataset

    def transform(self, dataset, inplace: bool = True):
        if not inplace:
            dataset = copy.deepcopy(dataset)
        with torch.no_grad():
            for i, data in enumerate(dataset):
                dataset[i] = self._postprocess(self._transform(self._preprocess(data)))
        return dataset

    def fit_transform(self, dataset, inplace: bool = True):
        if not inplace:
            dataset = copy.deepcopy(dataset)
        with torch.no_grad():
            for i, data in enumerate(dataset):
                dataset[i] = self._postprocess(self._transform(self._fit(self._preprocess(data))))
        return dataset

    def _preprocess(
            self, data: typing.Union[autogl.data.graph.GeneralStaticGraph, typing.Any]
    ) -> typing.Union[autogl.data.graph.GeneralStaticGraph, typing.Any]:
        return data

    def _postprocess(
            self, data: typing.Union[autogl.data.graph.GeneralStaticGraph, typing.Any]
    ) -> typing.Union[autogl.data.graph.GeneralStaticGraph, typing.Any]:
        return data

    def _fit(
            self, data: typing.Union[autogl.data.graph.GeneralStaticGraph, typing.Any]
    ) -> typing.Union[autogl.data.graph.GeneralStaticGraph, typing.Any]:
        return data

    def _transform(
            self, data: typing.Union[autogl.data.graph.GeneralStaticGraph, typing.Any]
    ) -> typing.Union[autogl.data.graph.GeneralStaticGraph, typing.Any]:
        return data



import typing as _typing
import autogl.data.graph


class _AbstractBaseFeatureEngineer:
    def fit(self, dataset):
        raise NotImplementedError

    def transform(self, dataset, inplace: bool = True):
        raise NotImplementedError

    def fit_transform(self, dataset, inplace: bool = True):
        raise NotImplementedError


class _ComposedFeatureEngineer(_AbstractBaseFeatureEngineer):
    @property
    def fe_components(self) -> _typing.Iterable[_AbstractBaseFeatureEngineer]:
        return self.__fe_components

    def fit(self, dataset):
        for fe in self.fe_components:
            dataset = fe.fit(dataset)
        return dataset

    def transform(self, dataset, inplace: bool = True):
        for fe in self.fe_components:
            dataset = fe.transform(dataset, inplace)
        return dataset

    def fit_transform(self, dataset, inplace: bool = True):
        for fe in self.fe_components:
            dataset = fe.fit(dataset)
        for fe in self.fe_components:
            dataset = fe.transform(dataset)
        return dataset

    def __init__(self, feature_engineers: _typing.Iterable[_AbstractBaseFeatureEngineer]):
        self.__fe_components: _typing.List[_AbstractBaseFeatureEngineer] = []
        for fe in feature_engineers:
            if isinstance(fe, _ComposedFeatureEngineer):
                self.__fe_components.extend(fe.fe_components)
            else:
                self.__fe_components.append(fe)

    def __and__(self, other: _AbstractBaseFeatureEngineer):
        return _ComposedFeatureEngineer((self, other))


class _BaseFeatureEngineer(_AbstractBaseFeatureEngineer):
    def __and__(self, other):
        return _ComposedFeatureEngineer((self, other))

    def fit(self, dataset):
        raise NotImplementedError

    def transform(self, dataset, inplace: bool = True):
        raise NotImplementedError

    def fit_transform(self, dataset, inplace: bool = True):
        return self.transform(self.fit(dataset), inplace)

    def _preprocess(
            self, data: _typing.Union[autogl.data.graph.GeneralStaticGraph, _typing.Any]
    ) -> _typing.Union[autogl.data.graph.GeneralStaticGraph, _typing.Any]:
        raise NotImplementedError

    def _postprocess(
            self, data: _typing.Union[autogl.data.graph.GeneralStaticGraph, _typing.Any]
    ) -> _typing.Union[autogl.data.graph.GeneralStaticGraph, _typing.Any]:
        raise NotImplementedError

    def _fit(
            self, data: _typing.Union[autogl.data.graph.GeneralStaticGraph, _typing.Any]
    ) -> _typing.Union[autogl.data.graph.GeneralStaticGraph, _typing.Any]:
        raise NotImplementedError

    def _transform(
            self, data: _typing.Union[autogl.data.graph.GeneralStaticGraph, _typing.Any]
    ) -> _typing.Union[autogl.data.graph.GeneralStaticGraph, _typing.Any]:
        raise NotImplementedError


class BaseFeatureEngineer(_BaseFeatureEngineer):
    def fit(self, dataset):
        raise NotImplementedError

    def transform(self, dataset, inplace: bool = True):
        raise NotImplementedError

    def _preprocess(
            self, data: _typing.Union[autogl.data.graph.GeneralStaticGraph, _typing.Any]
    ) -> _typing.Union[autogl.data.graph.GeneralStaticGraph, _typing.Any]:
        return data

    def _postprocess(
            self, data: _typing.Union[autogl.data.graph.GeneralStaticGraph, _typing.Any]
    ) -> _typing.Union[autogl.data.graph.GeneralStaticGraph, _typing.Any]:
        return data

    def _fit(
            self, data: _typing.Union[autogl.data.graph.GeneralStaticGraph, _typing.Any]
    ) -> _typing.Union[autogl.data.graph.GeneralStaticGraph, _typing.Any]:
        return data

    def _transform(
            self, data: _typing.Union[autogl.data.graph.GeneralStaticGraph, _typing.Any]
    ) -> _typing.Union[autogl.data.graph.GeneralStaticGraph, _typing.Any]:
        return data



import re

import torch


def index_to_mask(index, size):
    mask = torch.zeros(size, dtype=torch.bool, device=index.device)
    mask[index] = 1
    return mask


class Data(object):
    r"""A plain python object modeling a single graph with various
    (optional) attributes:

    Args:
        x (Tensor, optional): Node feature matrix with shape :obj:`[num_nodes,
            num_node_features]`. (default: :obj:`None`)
        edge_index (LongTensor, optional): Graph connectivity in COO format
            with shape :obj:`[2, num_edges]`. (default: :obj:`None`)
        edge_attr (Tensor, optional): Edge feature matrix with shape
            :obj:`[num_edges, num_edge_features]`. (default: :obj:`None`)
        y (Tensor, optional): Graph or node targets with arbitrary shape.
            (default: :obj:`None`)
        pos (Tensor, optional): Node position matrix with shape
            :obj:`[num_nodes, num_dimensions]`. (default: :obj:`None`)

    The data object is not restricted to these attributes and can be extented
    by any other additional data.
    """

    def __init__(self, x=None, edge_index=None, edge_attr=None, y=None, pos=None):
        self.x = x
        self.edge_index = edge_index
        self.edge_attr = edge_attr
        self.y = y
        self.pos = pos

    @staticmethod
    def from_dict(dictionary):
        r"""Creates a data object from a python dictionary."""
        data = Data()
        for key, item in dictionary.items():
            data[key] = item
        return data

    def __getitem__(self, key):
        r"""Gets the data of the attribute :obj:`key`."""
        return getattr(self, key)

    def __setitem__(self, key, value):
        """Sets the attribute :obj:`key` to :obj:`value`."""
        setattr(self, key, value)

    @property
    def keys(self):
        r"""Returns all names of graph attributes."""
        keys = [key for key in self.__dict__.keys() if self[key] is not None]
        keys = [key for key in keys if key[:2] != "__" and key[-2:] != "__"]
        return keys

    def __len__(self):
        r"""Returns the number of all present attributes."""
        return len(self.keys)

    def __contains__(self, key):
        r"""Returns :obj:`True`, if the attribute :obj:`key` is present in the
        data."""
        return key in self.keys

    def __iter__(self):
        r"""Iterates over all present attributes in the data, yielding their
        attribute names and content."""
        for key in sorted(self.keys):
            yield key, self[key]

    def __call__(self, *keys):
        r"""Iterates over all attributes :obj:`*keys` in the data, yielding
        their attribute names and content.
        If :obj:`*keys` is not given this method will iterative over all
        present attributes."""
        for key in sorted(self.keys) if not keys else keys:
            if self[key] is not None:
                yield key, self[key]

    def cat_dim(self, key, value):
        r"""Returns the dimension in which the attribute :obj:`key` with
        content :obj:`value` gets concatenated when creating batches.

        .. note::

            This method is for internal use only, and should only be overridden
            if the batch concatenation process is corrupted for a specific data
            attribute.
        """
        # `*index*` and `*face*` should be concatenated in the last dimension,
        # everything else in the first dimension.
        return -1 if bool(re.search("(index|face)", key)) else 0

    # own methods for processing
    def get_label_number(self):
        r"""Get the number of labels in this dataset as dict."""
        label_num = {}
        labels = self.y.unique().cpu().detach().numpy().tolist()
        for label in labels:
            label_num[label] = (self.y == label).sum().item()
        return label_num

    def random_splits_mask(self, train_ratio, val_ratio, seed=None):
        r"""If the data has masks for train/val/test, return the splits with specific ratio.

        Parameters
        ----------
        train_ratio : float
            the portion of data that used for training.

        val_ratio : float
            the portion of data that used for validation.

        seed : int
            random seed for splitting dataset.
        """
        rs = torch.get_rng_state()
        rs_cuda = torch.cuda.get_rng_state()
        if seed is not None:
            torch.manual_seed(seed)
            torch.cuda.manual_seed(seed)

        perm = torch.randperm(self.num_nodes)
        train_index = perm[: int(self.num_nodes * train_ratio)]
        val_index = perm[
            int(self.num_nodes * train_ratio) : int(
                self.num_nodes * (train_ratio + val_ratio)
            )
        ]
        test_index = perm[int(self.num_nodes * (train_ratio + val_ratio)) :]
        self.train_mask = index_to_mask(train_index, size=self.num_nodes)
        self.val_mask = index_to_mask(val_index, size=self.num_nodes)
        self.test_mask = index_to_mask(test_index, size=self.num_nodes)

        torch.set_rng_state(rs)
        torch.cuda.set_rng_state(rs_cuda)
        return self

    def random_splits_nodes(self, train_ratio, val_ratio, seed=None):
        r"""If the data uses id of nodes for train/val/test, return the splits with specific ratio.

        Parameters
        ----------
        train_ratio : float
            the portion of data that used for training.

        val_ratio : float
            the portion of data that used for validation.

        seed : int
            random seed for splitting dataset.
        """
        rs = torch.get_rng_state()
        rs_cuda = torch.cuda.get_rng_state()
        if seed is not None:
            torch.manual_seed(seed)
            torch.cuda.manual_seed(seed)

        perm = torch.randperm(self.num_nodes)
        self.train_node = perm[: int(self.num_nodes * train_ratio)]
        self.val_node = perm[
            int(self.num_nodes * train_ratio) : int(
                self.num_nodes * (train_ratio + val_ratio)
            )
        ]
        self.test_node = perm[int(self.num_nodes * (train_ratio + val_ratio)) :]
        self.train_target = self.y[self.train_node]
        self.valid_target = self.y[self.valid_node]
        self.test_target = self.y[self.test_node]

        torch.set_rng_state(rs)
        torch.cuda.set_rng_state(rs_cuda)
        return self

    def random_splits_mask_class(
        self, num_train_per_class, num_val, num_test, seed=None
    ):
        r"""If the data has masks for train/val/test, return the splits with specific number of samples from every class for training.

        Parameters
        ----------
        num_train_per_class : int
            the number of samples from every class used for training.

        num_val : int
            the total number of nodes that used for validation.

        num_test : int
            the total number of nodes that used for testing.

        seed : int
            random seed for splitting dataset.
        """
        rs = torch.get_rng_state()
        rs_cuda = torch.cuda.get_rng_state()
        if seed is not None:
            torch.manual_seed(seed)
            torch.cuda.manual_seed(seed)

        num_classes = self.y.max().cpu().item() + 1
        self.train_mask.fill_(False)
        for c in range(num_classes):
            idx = (self.y == c).nonzero().view(-1)
            idx = idx[torch.randperm(idx.size(0))[:num_train_per_class]]
            self.train_mask[idx] = True

        remaining = (~self.train_mask).nonzero().view(-1)
        remaining = remaining[torch.randperm(remaining.size(0))]

        self.val_mask.fill_(False)
        self.val_mask[remaining[:num_val]] = True

        self.test_mask.fill_(False)
        self.test_mask[remaining[num_val : num_val + num_test]] = True

        torch.set_rng_state(rs)
        torch.cuda.set_rng_state(rs_cuda)
        return self

    def random_splits_nodes_class(
        self, num_train_per_class, num_val, num_test, seed=None
    ):
        r"""If the data uses id of nodes for train/val/test, return the splits with specific number of samples from every class for training.

        Parameters
        ----------
        num_train_per_class : int
            the number of samples from every class used for training.

        num_val : int
            the total number of nodes that used for validation.

        num_test : int
            the total number of nodes that used for testing.

        seed : int
            random seed for splitting dataset.
        """
        rs = torch.get_rng_state()
        rs_cuda = torch.cuda.get_rng_state()
        if seed is not None:
            torch.manual_seed(seed)
            torch.cuda.manual_seed(seed)

        num_classes = self.y.max().cpu().item() + 1
        train_mask = torch.zeros(
            self.num_nodes, dtype=torch.bool, device=self.train_node.device
        )
        sup = []
        for c in range(num_classes):
            idx = (self.y == c).nonzero().view(-1)
            idx = idx[torch.randperm(idx.size(0))[:num_train_per_class]]
            sup.append(idx)
            train_mask[idx] = True

        self.train_node = torch.cat(sup)

        remaining = (~train_mask).nonzero().view(-1)
        remaining = remaining[torch.randperm(remaining.size(0))]

        self.val_node = remaining[:num_val]
        self.test_node = remaining[num_val : num_val + num_test]

        self.train_target = self.y[self.train_node]
        self.valid_target = self.y[self.valid_node]
        self.test_target = self.y[self.test_node]

        torch.set_rng_state(rs)
        torch.cuda.set_rng_state(rs_cuda)
        return self

    def __inc__(self, key, value):
        r""" "Returns the incremental count to cumulatively increase the value
        of the next attribute of :obj:`key` when creating batches.

        .. note::

            This method is for internal use only, and should only be overridden
            if the batch concatenation process is corrupted for a specific data
            attribute.
        """
        # Only `*index*` and `*face*` should be cumulatively summed up when
        # creating batches.
        return self.num_nodes if bool(re.search("(index|face)", key)) else 0

    @property
    def num_edges(self):
        r"""Returns the number of edges in the graph."""
        for key, item in self("edge_index", "edge_attr"):
            return item.size(self.cat_dim(key, item))
        return None

    @property
    def num_features(self):
        r"""Returns the number of features per node in the graph."""
        return 1 if self.x.dim() == 1 else self.x.size(1)

    @property
    def num_nodes(self):
        if self.x is not None:
            return self.x.shape[0]
        return torch.max(self.edge_index) + 1

    def is_coalesced(self):
        r"""Returns :obj:`True`, if edge indices are ordered and do not contain
        duplicate entries."""
        row, col = self.edge_index
        index = self.num_nodes * row + col
        return row.size(0) == torch.unique(index).size(0)

    def apply(self, func, *keys):
        r"""Applies the function :obj:`func` to all attributes :obj:`*keys`.
        If :obj:`*keys` is not given, :obj:`func` is applied to all present
        attributes.
        """
        for key, item in self(*keys):
            self[key] = func(item)
        return self

    def contiguous(self, *keys):
        r"""Ensures a contiguous memory layout for all attributes :obj:`*keys`.
        If :obj:`*keys` is not given, all present attributes are ensured to
        have a contiguous memory layout."""
        return self.apply(lambda x: x.contiguous(), *keys)

    def to(self, device, *keys):
        r"""Performs tensor dtype and/or device conversion to all attributes
        :obj:`*keys`.
        If :obj:`*keys` is not given, the conversion is applied to all present
        attributes."""
        return self.apply(lambda x: x.to(device), *keys)

    def cuda(self, *keys):
        return self.apply(lambda x: x.cuda(), *keys)

    def clone(self):
        return Data.from_dict({k: v.clone() for k, v in self})

    def __repr__(self):
        info = [
            "{}={}".format(key, list(item.size()))
            for key, item in self
            if type(item) != list and type(item) != dict
        ]
        return "{}({})".format(self.__class__.__name__, ", ".join(info))



from .base import BaseSelector
import numpy as np
from .. import register_feature


@register_feature("SeFilterConstant")
class SeFilterConstant(BaseSelector):
    r"""drop constant features"""

    def _fit(self, data):
        d1, d2 = data.x.shape
        xx = data.x
        # if d2>=d1:
        #     if np.allclose(xx[:,:d1],np.eye(d1)):
        #         return np.empty((d1,0))
        self._sel = np.where(np.all(xx == xx[0, :], axis=0) == False)[0]



import torch
import typing as _typing
from . import _canonical_edge_type


class SpecificTypedNodeDataView(_typing.MutableMapping[str, torch.Tensor]):
    def __getitem__(self, data_key: str) -> torch.Tensor:
        raise NotImplementedError

    def __setitem__(self, data_key: str, value: torch.Tensor):
        raise NotImplementedError

    def __delitem__(self, data_key: str) -> None:
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError

    def __iter__(self) -> _typing.Iterator[str]:
        raise NotImplementedError


class SpecificTypedNodeView:
    @property
    def data(self) -> SpecificTypedNodeDataView:
        raise NotImplementedError

    @data.setter
    def data(self, nodes_data: _typing.Mapping[str, torch.Tensor]):
        raise NotImplementedError


class HeterogeneousNodeView(_typing.Iterable[str]):
    @property
    def data(self) -> SpecificTypedNodeDataView:
        raise NotImplementedError

    @data.setter
    def data(self, nodes_data: _typing.Mapping[str, torch.Tensor]):
        raise NotImplementedError

    def __getitem__(self, node_type: _typing.Optional[str]) -> SpecificTypedNodeView:
        raise NotImplementedError

    def __setitem__(
            self, node_t: _typing.Optional[str],
            nodes_data: _typing.Mapping[str, torch.Tensor]
    ):
        raise NotImplementedError

    def __delitem__(self, node_t: _typing.Optional[str]):
        raise NotImplementedError

    def __iter__(self) -> _typing.Iterator[str]:
        raise NotImplementedError

    @property
    def is_homogeneous(self) -> bool:
        raise NotImplementedError


class HomogeneousEdgesDataView(_typing.MutableMapping[str, torch.Tensor]):
    def __getitem__(self, data_key: str) -> torch.Tensor:
        raise NotImplementedError

    def __setitem__(self, data_key: str, value: torch.Tensor):
        raise NotImplementedError

    def __delitem__(self, data_key: str):
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError

    def __iter__(self) -> _typing.Iterator[str]:
        raise NotImplementedError


class HomogeneousEdgesView:
    @property
    def connections(self) -> torch.LongTensor:
        raise NotImplementedError

    @property
    def data(self) -> HomogeneousEdgesDataView:
        raise NotImplementedError


class HeterogeneousEdgesView(_typing.Collection[_canonical_edge_type.CanonicalEdgeType]):
    @property
    def connections(self) -> torch.LongTensor:
        raise NotImplementedError

    @property
    def data(self) -> HomogeneousEdgesDataView:
        raise NotImplementedError

    @property
    def is_homogeneous(self) -> bool:
        raise NotImplementedError

    def set(
            self, edge_t: _typing.Union[None, str, _typing.Tuple[str, str, str]],
            connections: torch.LongTensor, data: _typing.Optional[_typing.Mapping[str, torch.Tensor]] = ...
    ):
        raise NotImplementedError

    def __getitem__(
            self,
            edge_t: _typing.Union[
                None, str, _typing.Tuple[str, str, str], _canonical_edge_type.CanonicalEdgeType
            ]
    ) -> HomogeneousEdgesView:
        raise NotImplementedError

    def __setitem__(
            self,
            edge_t: _typing.Union[
                None, str, _typing.Tuple[str, str, str], _canonical_edge_type.CanonicalEdgeType
            ],
            edges: _typing.Union[torch.LongTensor]
    ):
        raise NotImplementedError

    def __delitem__(
            self,
            edge_t: _typing.Union[
                None, str, _typing.Tuple[str, str, str], _canonical_edge_type.CanonicalEdgeType
            ]
    ):
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError

    def __iter__(self) -> _typing.Iterator[_canonical_edge_type.CanonicalEdgeType]:
        raise NotImplementedError

    def __contains__(
            self,
            edge_type: _typing.Union[
                str, _typing.Tuple[str, str, str], _canonical_edge_type.CanonicalEdgeType
            ]
    ) -> bool:
        raise NotImplementedError


class GraphDataView(_typing.MutableMapping[str, torch.Tensor]):
    def __setitem__(self, data_key: str, data: torch.Tensor) -> None:
        raise NotImplementedError

    def __delitem__(self, data_key: str) -> None:
        raise NotImplementedError

    def __getitem__(self, data_key: str) -> torch.Tensor:
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError

    def __iter__(self) -> _typing.Iterator[str]:
        raise NotImplementedError



from .message_passing import MessagePassing
from .gcn_conv import GCNConv
from .cheb_conv import ChebConv
from .sage_conv import SAGEConv
from .gat_conv import GATConv
from .gin_conv import GINConv, GINEConv
from .arma_conv import ARMAConv
from .edge_conv import EdgeConv, DynamicEdgeConv

__all__ = [
    'MessagePassing',
    'GCNConv',
    'ChebConv',
    'SAGEConv',
    'GATConv',
    'GINConv',
    'GINEConv',
    'ARMAConv',
    'EdgeConv',
    'DynamicEdgeConv',
]

classes = __all__



from copy import deepcopy
from ..base import BaseAutoModel

class BaseHeteroModelMaintainer(BaseAutoModel):
    def __init__(self, num_features, num_classes, device, dataset=None, **kwargs):
        super().__init__(num_features, num_classes, device, **kwargs)
        self._registered_parameters = {}
        if dataset is not None:
            self.from_dataset(dataset)

    def from_dataset(self, dataset):
        raise NotImplementedError

    # consider moving this to inner classes
    def register_parameter(self, key: str, value):
        self._registered_parameters[key] = value
        setattr(self, key, value)

    def destroy_parameter(self, key):
        if key in self._registered_parameters:
            return self._registered_parameters.pop(key)
        return None

    def from_hyper_parameter(self, hp, **kwargs):
        kw = deepcopy(self._kwargs)
        kw.update(kwargs)
        ret_self = self.__class__(
            self.input_dimension,
            self.output_dimension,
            self.device,
            **kw
        )
        hp_now = dict(self.hyper_parameters)
        hp_now.update(hp)
        ret_self.hyper_parameters = hp_now
        for key, value in self._registered_parameters.items():
            ret_self.register_parameter(key, value)
        ret_self.initialize()
        return ret_self



from scipy.sparse import csr_matrix
import scipy.sparse as ssp
import networkx as nx
from .base import BaseGenerator
import numpy as np
from .. import register_feature


class Eigen(object):
    def __init__(self):
        pass

    def normalize_adj(self, adj):
        rowsum = np.array(adj.sum(1))
        d_inv_sqrt = np.power(rowsum, -0.5).flatten()
        d_inv_sqrt[np.isinf(d_inv_sqrt)] = 0.0
        d_inv_sqrt = ssp.diags(d_inv_sqrt)
        return adj.dot(d_inv_sqrt).transpose().dot(d_inv_sqrt)

    def forward(self, adj, d, use_eigenvalues=0, adj_norm=1):
        G = nx.from_scipy_sparse_matrix(adj)
        comp = list(nx.connected_components(G))
        results = np.zeros((adj.shape[0], d))
        for i in range(len(comp)):
            node_index = np.array(list(comp[i]))
            d_temp = min(len(node_index) - 2, d)
            if d_temp <= 0:
                continue
            temp_adj = adj[node_index, :][:, node_index].asfptype()
            if adj_norm == 1:
                temp_adj = self.normalize_adj(temp_adj)
            lamb, X = ssp.linalg.eigs(temp_adj, d_temp)
            lamb, X = lamb.real, X.real
            temp_order = np.argsort(lamb)
            lamb, X = lamb[temp_order], X[:, temp_order]
            for i in range(X.shape[1]):
                if np.sum(X[:, i]) < 0:
                    X[:, i] = -X[:, i]
            if use_eigenvalues == 1:
                X = X.dot(np.diag(np.sqrt(np.absolute(lamb))))
            elif use_eigenvalues == 2:
                X = X.dot(np.diag(lamb))
            results[node_index, :d_temp] = X
        return results


@register_feature("eigen")
class GeEigen(BaseGenerator):
    r"""concat Eigen features

    Notes
    -----
    An implementation of [#]_

    References
    ----------
    .. [#] Ziwei Zhang, Peng Cui, Jian Pei, Xin Wang, Wenwu Zhu:
        Eigen-GNN: A Graph Structure Preserving Plug-in for GNNs. CoRR abs/2006.04330 (2020)
        https://arxiv.org/abs/2006.04330


    Parameters
    ----------
    size : int
        EigenGNN hidden size
    """

    def __init__(self, size=32):
        super(GeEigen, self).__init__()
        self.size = size

    def _transform(self, data):

        adj = csr_matrix(
            (data.edge_weight, (data.edge_index[0, :], data.edge_index[1, :])),
            shape=(data.x.shape[0], data.x.shape[0]),
        )
        if np.max(adj - adj.T) > 1e-5:
            adj = adj + adj.T
        mf = Eigen()
        data.x = np.concatenate([data.x, mf.forward(adj, self.size)], axis=1)

        return data

    def _preprocess(self, data):
        if not hasattr(data, "edge_weight"):
            data.edge_weight = np.ones(data.edge_index.shape[1])

    def _postprocess(self, data):
        del data.edge_weight



import logging
import torch.nn.functional
import typing as _typing
import torch_geometric
from torch_geometric.nn.conv import GATConv
from .. import base_encoder, encoder_registry
from ... import _utils


class GATUtils:
    @classmethod
    def to_total_hidden_dimensions(
            cls, per_head_output_dimensions: _typing.Sequence[int],
            num_hidden_heads: int, num_output_heads: int, concat_last: bool = False,
    ) -> _typing.Sequence[int]:
        return [
            d * (num_hidden_heads if layer < (len(per_head_output_dimensions) - 1) else (num_output_heads if concat_last else 1))
            for layer, d in enumerate(per_head_output_dimensions)
        ]


class _GAT(torch.nn.Module):
    def __init__(
            self, input_dimension: int,
            per_head_output_dimensions: _typing.Sequence[int],
            num_hidden_heads: int, num_output_heads: int,
            _dropout: float, _act: _typing.Optional[str],
            concat_last: bool = True
    ):
        super(_GAT, self).__init__()
        self._dropout: float = _dropout
        self._act: _typing.Optional[str] = _act
        total_output_dimensions: _typing.Sequence[int] = (
            GATUtils.to_total_hidden_dimensions(
                per_head_output_dimensions, num_hidden_heads, num_output_heads, concat_last = concat_last
            )
        )
        num_layers = len(per_head_output_dimensions)
        self.__convolution_layers: torch.nn.ModuleList = torch.nn.ModuleList()
        for layer in range(len(per_head_output_dimensions)):
            self.__convolution_layers.append(
                GATConv(
                    input_dimension if layer == 0 else total_output_dimensions[layer - 1],
                    per_head_output_dimensions[layer],
                    num_hidden_heads if layer < num_layers - 1 else num_output_heads,
                    dropout=_dropout,
                    concat=True if layer < num_layers - 1 or concat_last else False
                )
            )

    def forward(self, data: torch_geometric.data.Data, *__args, **__kwargs):
        x: torch.Tensor = data.x
        edge_index: torch.LongTensor = data.edge_index
        if (
                hasattr(data, "edge_weight") and
                isinstance(getattr(data, "edge_weight"), torch.Tensor)
                and torch.is_tensor(data.edge_weight)
        ):
            edge_weight: _typing.Optional[torch.Tensor] = data.edge_weight
        else:
            edge_weight: _typing.Optional[torch.Tensor] = None
        results: _typing.MutableSequence[torch.Tensor] = [x]
        for layer, _gat in enumerate(self.__convolution_layers):
            x: torch.Tensor = torch.nn.functional.dropout(
                x, self._dropout, self.training
            )
            x: torch.Tensor = _gat(x, edge_index, edge_weight)
            if layer < len(self.__convolution_layers) - 1:
                x: torch.Tensor = _utils.activation.activation_func(x, self._act)
            results.append(x)
        logging.debug("{:d} layer, each layer shape {:s}".format(len(results), " ".join([str(x.shape) for x in results])))
        return results


@encoder_registry.EncoderUniversalRegistry.register_encoder('gat')
@encoder_registry.EncoderUniversalRegistry.register_encoder('gat_encoder')
class GATEncoderMaintainer(base_encoder.AutoHomogeneousEncoderMaintainer):
    def _initialize(self) -> _typing.Optional[bool]:
        dimensions = list(self.hyper_parameters['hidden'])
        concat_last = True
        if (
                self.final_dimension not in (Ellipsis, None)
                and isinstance(self.final_dimension, int)
                and self.final_dimension > 0
        ):
            dimensions.append(self.final_dimension)
            concat_last = False
        logging.debug("current dimensions %s", dimensions)
        self._encoder = _GAT(
            self.input_dimension,
            dimensions,
            self.hyper_parameters.get('num_hidden_heads'),
            self.hyper_parameters.get('num_output_heads'),
            self.hyper_parameters['dropout'],
            self.hyper_parameters['act'],
            concat_last=concat_last
        )
        return True

    def get_output_dimensions(self) -> _typing.Iterable[int]:
        temp = list(self.hyper_parameters["hidden"])
        concat_last = True
        if (
                self.final_dimension not in (Ellipsis, None) and
                isinstance(self.final_dimension, int) and
                self.final_dimension > 0
        ):
            temp.append(self.final_dimension)
            concat_last = False
        output_dimensions = [self.input_dimension]
        output_dimensions.extend(
            GATUtils.to_total_hidden_dimensions(
                temp,
                self.hyper_parameters.get("num_hidden_heads", self.hyper_parameters["heads"]),
                self.hyper_parameters.get("num_output_heads", 1),
                concat_last = concat_last
            )
        )
        return output_dimensions

    def __init__(
            self,
            input_dimension: _typing.Optional[int] = ...,
            final_dimension: _typing.Optional[int] = ...,
            device: _typing.Union[torch.device, str, int, None] = ...,
            *args, **kwargs
    ):
        super(GATEncoderMaintainer, self).__init__(
            input_dimension, final_dimension, device, *args, **kwargs
        )
        self.hyper_parameter_space = [
            {
                "parameterName": "num_layers",
                "type": "DISCRETE",
                "feasiblePoints": "2,3,4",
            },
            {
                "parameterName": "hidden",
                "type": "NUMERICAL_LIST",
                "numericalType": "INTEGER",
                "length": 3,
                "minValue": [8, 8, 8],
                "maxValue": [64, 64, 64],
                "scalingType": "LOG",
                "cutPara": ("num_layers",),
                "cutFunc": lambda x: x[0] - 1,
            },
            {
                "parameterName": "dropout",
                "type": "DOUBLE",
                "maxValue": 0.8,
                "minValue": 0.2,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "num_hidden_heads",
                "type": "DISCRETE",
                "feasiblePoints": "2,4,8,16",
            },
            {
                "parameterName": "num_output_heads",
                "type": "DISCRETE",
                "feasiblePoints": "2,4,8,16",
            },
            {
                "parameterName": "act",
                "type": "CATEGORICAL",
                "feasiblePoints": ["leaky_relu", "relu", "elu", "tanh"],
            },
        ]
        self.hyper_parameters = {
            "num_layers": 2,
            "hidden": [32],
            "num_hidden_heads": 4,
            "num_output_heads": 4,
            "dropout": 0.2,
            "act": "leaky_relu",
        }



import torch.nn as nn
import dgl.function as fn

from .. import register_model
from .base import BaseHeteroModelMaintainer
from ..base import activate_func
from .....utils import get_logger

LOGGER = get_logger("HGTModel")

def set_default(args, d):
    for k, v in d.items():
        if k not in args:
            args[k] = v
    return args

class HeteroRGCNLayer(nn.Module):
    def __init__(self, in_size, out_size, etypes):
        super(HeteroRGCNLayer, self).__init__()
        # W_r for each relation
        self.weight = nn.ModuleDict({
                name : nn.Linear(in_size, out_size) for name in etypes
            })

    def forward(self, G, feat_dict):
        # The input is a dictionary of node features for each type
        funcs = {}
        for srctype, etype, dsttype in G.canonical_etypes:
            # Compute W_r * h
            Wh = self.weight[etype](feat_dict[srctype])
            # Save it in graph for message passing
            G.nodes[srctype].data['Wh_%s' % etype] = Wh
            # Specify per-relation message passing functions: (message_func, reduce_func).
            # Note that the results are saved to the same destination feature 'h', which
            # hints the type wise reducer for aggregation.
            funcs[etype] = (fn.copy_u('Wh_%s' % etype, 'm'), fn.mean('m', 'h'))
        # Trigger message passing of multiple types.
        # The first argument is the message passing functions for each relation.
        # The second one is the type wise reducer, could be "sum", "max",
        # "min", "mean", "stack"
        G.multi_update_all(funcs, 'sum')
        # return the updated node feature dictionary
        return {ntype : G.nodes[ntype].data['h'] for ntype in G.ntypes}

class HeteroRGCN(nn.Module):
    def __init__(self, args):

        super(HeteroRGCN, self).__init__()
        self.args = args 
        self.edge_type = self.args["edge_type"]
        self.out_key = self.args["out_key"]
        missing_keys = list(
            set(
                [
                    "features_num",
                    "num_class",
                    "num_layers",
                    "hidden",
                    "act",
                    "out_key",
                    "edge_type"
                ]
            )
            - set(self.args.keys())
        )
        if len(missing_keys) > 0:
            raise Exception("Missing keys: %s." % ",".join(missing_keys))

        self.num_layers = int(self.args["num_layers"])

        if not self.num_layers == len(self.args["hidden"])-1:
            LOGGER.warn("Warning: layer size does not match the length of hidden units")

        self.layers  = nn.ModuleList()
        self.layers.append(HeteroRGCNLayer(self.args["features_num"], self.args["hidden"][0], self.edge_type))
        for i in range(self.num_layers-2):
            self.layers.append(HeteroRGCNLayer(self.args["hidden"][i], self.args["hidden"][i+1], self.edge_type))
        self.layers.append(HeteroRGCNLayer(self.args["hidden"][-1], self.args["num_class"], self.edge_type))
            
    def forward(self, G):
        h_dict = {ntype : G.nodes[ntype].data['feat'] for ntype in G.ntypes}
        for l in range(self.num_layers):
            h_dict = self.layers[l](G, h_dict)
            if l!=self.num_layers-1:
                h_dict = {k : activate_func(h, self.args["act"]) for k, h in h_dict.items()}    

        return h_dict[self.out_key]

@register_model("HeteroRGCN")
class AutoHeteroRGCN(BaseHeteroModelMaintainer):
    r"""
    AutoHeteroRGCN.
    The model used in this automodel is HeteroRGCN, i.e., the relational graph convolutional network from the
    `"Modeling Relational Data with Graph Convolutional Networks"
    <https://arxiv.org/abs/1703.06103>`_ paper

        
    Parameters
    ----------
    num_features: `int`.
        The dimension of features.

    num_classes: `int`.
        The number of classes.

    device: `torch.device` or `str`.
        The device where model will be running on.

    init: `bool`.
        If True(False), the model will (not) be initialized.

    dataset: `autogl.datasets`.
        Hetero Graph Dataset in autogl.
    """
    def __init__(
        self, num_features=None, num_classes=None, device=None, init=False, dataset=None, **args
    ):
        super().__init__(num_features, num_classes, device, dataset)
        
        self.hyper_parameter_space = [
            {
                "parameterName": "num_layers",
                "type": "DISCRETE",
                "feasiblePoints": "2,3,4",
            },
            {
                "parameterName": "hidden",
                "type": "NUMERICAL_LIST",
                "numericalType": "INTEGER",
                "length": 3,
                "minValue": [8, 8, 8],
                "maxValue": [64, 64, 64],
                "scalingType": "LOG",
                "cutPara": ("num_layers",),
                "cutFunc": lambda x: x[0] - 1,
            },
            {
                "parameterName": "dropout",
                "type": "DOUBLE",
                "maxValue": 0.8,
                "minValue": 0.2,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "act",
                "type": "CATEGORICAL",
                "feasiblePoints": ["leaky_relu", "relu", "elu", "tanh"],  
            },
        ]

        self.hyper_parameters = {
            "num_layers": 2,
            "hidden": [256],
            "act": "leaky_rely",
        }

        if init is True:
            self.initialize()

    def _initialize(self):
        # """Initialize model."""
        self._model = HeteroRGCN(dict(
            features_num=self.input_dimension,
            num_class=self.output_dimension,
            edge_type=self.edge_type,
            out_key=self.out_key,
            **self.hyper_parameters
        )).to(self.device)

    def from_dataset(self, dataset):
        self.register_parameter("out_key", dataset.schema['target_node_type'])
        self.register_parameter("edge_type", dataset[0].etypes)



import autogl

if autogl.backend.DependentBackend.is_dgl():
    from ._base_feature_engineer_dgl import BaseFeatureEngineer
else:
    from ._base_feature_engineer_pyg import BaseFeatureEngineer


class BaseFeature(BaseFeatureEngineer):
    ...



from .base import BaseEnsembler

ENSEMBLE_DICT = {}


def register_ensembler(name):
    def register_ensembler_cls(cls):
        if name in ENSEMBLE_DICT:
            raise ValueError("Cannot register duplicate ensembler ({})".format(name))
        if not issubclass(cls, BaseEnsembler):
            raise ValueError(
                "Model ({}: {}) must extend BaseEnsembler".format(name, cls.__name__)
            )
        ENSEMBLE_DICT[name] = cls
        return cls

    return register_ensembler_cls


from .voting import Voting
from .stacking import Stacking


def build_ensembler_from_name(name: str) -> BaseEnsembler:
    """
    Parameters
    ----------
    name: ``str``
        the name of ensemble module.

    Returns
    -------
    BaseEnsembler:
        the ensembler built using default parameters

    Raises
    ------
    AssertionError
        If an invalid name is passed in
    """
    assert name in ENSEMBLE_DICT, "ensemble module do not have name " + name
    return ENSEMBLE_DICT[name]()


__all__ = ["BaseEnsembler", "Voting", "Stacking", "build_ensembler_from_name"]



import logging
import dgl
import math
import torch
import torch.nn as nn
import torch.nn.functional as F
import dgl.function as fn
from dgl.nn.functional import edge_softmax
from dgl.nn.pytorch import GATConv
from functools import partial

from ..._utils import activation
from .. import register_model
from .base import BaseHeteroModelMaintainer
from .....utils import get_logger

LOGGER = get_logger("HANModel")

def set_default(args, d):
    for k, v in d.items():
        if k not in args:
            args[k] = v
    return args

class SemanticAttention(nn.Module):
    def __init__(self, in_size, hidden_size=128):
        super(SemanticAttention, self).__init__()

        self.project = nn.Sequential(
            nn.Linear(in_size, hidden_size),
            nn.Tanh(),
            nn.Linear(hidden_size, 1, bias=False)
        )

    def forward(self, z):
        w = self.project(z).mean(0)                    # (M, 1)
        beta = torch.softmax(w, dim=0)                 # (M, 1)
        beta = beta.expand((z.shape[0],) + beta.shape) # (N, M, 1)

        return (beta * z).sum(1)                       # (N, D * K)

class HANLayer(nn.Module):
    """
    HAN layer.

    Arguments
    ---------
    meta_paths : list of metapaths, each as a list of edge types
    in_size : input feature dimension
    out_size : output feature dimension
    layer_num_heads : number of attention heads
    dropout : Dropout probability

    Inputs
    ------
    g : DGLHeteroGraph
        The heterogeneous graph
    h : tensor
        Input features

    Outputs
    -------
    tensor
        The output feature
    """
    def __init__(self, meta_paths, in_size, out_size, layer_num_heads, dropout, act):
        super(HANLayer, self).__init__()

        # One GAT layer for each meta path based adjacency matrix
        self.gat_layers = nn.ModuleList()
        for i in range(len(meta_paths)):
            self.gat_layers.append(GATConv(in_size, out_size, layer_num_heads,
                                           dropout, dropout, activation=act,
                                           allow_zero_in_degree=True))
        self.semantic_attention = SemanticAttention(in_size=out_size * layer_num_heads)
        self.meta_paths = list(tuple(meta_path) for meta_path in meta_paths)

        self._cached_graph = None
        self._cached_coalesced_graph = {}

    def forward(self, g, h):
        semantic_embeddings = []

        if self._cached_graph is None or self._cached_graph is not g:
            self._cached_graph = g
            self._cached_coalesced_graph.clear()
            for meta_path in self.meta_paths:
                self._cached_coalesced_graph[meta_path] = dgl.metapath_reachable_graph(
                        g, meta_path)

        for i, meta_path in enumerate(self.meta_paths):
            new_g = self._cached_coalesced_graph[meta_path]
            semantic_embeddings.append(self.gat_layers[i](new_g, h).flatten(1))
        semantic_embeddings = torch.stack(semantic_embeddings, dim=1)                  # (N, M, D * K)

        return self.semantic_attention(semantic_embeddings)                            # (N, D * K)


class HAN(nn.Module):
    def __init__(self, args):
        super(HAN, self).__init__()
        self.args = args
        missing_keys = list(
            set(
                [
                    "meta_paths",
                    "num_features",
                    "num_class",
                    "out_key",
                    "num_layers",
                    "hidden",
                    "heads",
                    "dropout",
                    "act",
                ]
            )
            - set(self.args.keys())
        )
        if len(missing_keys) > 0:
            raise Exception("Missing keys: %s." % ",".join(missing_keys))

        act = partial(activation.activation_func, function_name=self.args.get("act", "relu"))
        self.layers = nn.ModuleList()
        self.layers.append(HANLayer(self.args["meta_paths"], self.args["num_features"], self.args["hidden"][0], self.args["heads"][0], self.args["dropout"], act))
        for l in range(1, len(self.args["hidden"])):
            self.layers.append(HANLayer(self.args["meta_paths"], self.args["hidden"][l-1] * self.args["heads"][l-1],
                                        self.args["hidden"][l], self.args["heads"][l], self.args["dropout"], act))
        self.predict = nn.Linear(self.args["hidden"][-1] * self.args["heads"][-1], self.args["num_class"])

    def forward(self, g):
        h = g.nodes[self.args["out_key"]].data['feat']
        for gnn in self.layers:
            h = gnn(g, h)

        return self.predict(h)

@register_model("han")
class AutoHAN(BaseHeteroModelMaintainer):
    r"""
    AutoHAN.
    The model used in this automodel is HAN, i.e., the graph convolutional network from the
    `"Heterogenous Graph Attention Network"
    <https://arxiv.org/pdf/1903.07293.pdf>`_ paper.


    Parameters
    ----------
    num_features: `int`.
        The dimension of features.

    num_classes: `int`.
        The number of classes.

    device: `torch.device` or `str`.
        The device where model will be running on.

    init: `bool`.
        If True(False), the model will (not) be initialized.

    dataset: `autogl.datasets`.
        Hetero Graph Dataset in autogl.
    """
    def __init__(
        self, num_features=None, num_classes=None, device=None, init=False, dataset=None, **args
    ):
        super().__init__(num_features, num_classes, device, dataset, **args)

        self.hyper_parameter_space = [
            {
                "parameterName": "num_layers",
                "type": "DISCRETE",
                "feasiblePoints": "2,3,4",
            },
            {
                "parameterName": "hidden",
                "type": "NUMERICAL_LIST",
                "numericalType": "INTEGER",
                "length": 3,
                "minValue": [8, 8, 8],
                "maxValue": [64, 64, 64],
                "scalingType": "LOG",
                "cutPara": ("num_layers",),
                "cutFunc": lambda x: x[0] - 1,
            },
            {
                "parameterName": "dropout",
                "type": "DOUBLE",
                "maxValue": 0.8,
                "minValue": 0.2,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "heads",
                "type": "NUMERICAL_LIST",
                "numericalType": "INTEGER",
                "scalingType": "LOG",
                "length": 3,
                "minValue": [1, 1, 1],
                "maxValue": [16, 16, 16],
                "cutPara": ("num_layers",),
                "cutFunc": lambda x: x[0] - 1,
            },
            {
                "parameterName": "act",
                "type": "CATEGORICAL",
                "feasiblePoints": ["leaky_relu", "relu", "elu", "tanh","gelu"],  # add F.gelu()
            },
        ]

        self.hyper_parameters = {
            "num_layers": 2,
            "hidden": [256],
            "heads": [8],
            "dropout": 0.2,
            "act": "gelu",
        }

        if init is True:
            self.initialize()

    def _initialize(self):
        self._model = HAN(dict(
            num_features=self.input_dimension,
            num_class=self.output_dimension,
            out_key=self.out_key,
            meta_paths=self.meta_paths,
            **self.hyper_parameters
        )).to(self.device)

    def from_dataset(self, dataset):
        self.register_parameter("out_key", dataset.schema["target_node_type"])
        self.register_parameter("meta_paths", dataset.schema.meta_paths)



# tackle different backend
from autogl.backend import DependentBackend
_isdgl=DependentBackend.is_dgl()

def is_dgl():
    return _isdgl

if is_dgl():
    def bk_mask(data,mask):
        return data.ndata[f'{mask}_mask']
    def bk_label(data):
        return data.ndata['label']
    def bk_feat(data):
        return data.ndata['feat']
    def bk_gconv(op,data,feat):
        return op(data,feat)
else:
    def bk_mask(data,mask):
        return data[f'{mask}_mask']
    def bk_label(data):
        return data.y
    def bk_feat(data):
        return data.x
    def bk_gconv(op,data,feat):
        return op(feat,data.edge_index)



"""
Node classification Het Trainer Implementation
"""

from . import register_trainer

from .base import BaseNodeClassificationHetTrainer, EarlyStopping
import torch
from torch.optim.lr_scheduler import (
    StepLR,
    MultiStepLR,
    ExponentialLR,
    ReduceLROnPlateau,
)
import torch.nn.functional as F
from ..model import BaseAutoModel
from .evaluation import get_feval, Logloss
from typing import Union
from copy import deepcopy
from sklearn.metrics import f1_score

from ...utils import get_logger

from ...backend import DependentBackend

LOGGER = get_logger("node classification het trainer")

def score(logits, labels):
    _, indices = torch.max(logits, dim=1)
    prediction = indices.long().cpu().numpy()
    labels = labels.cpu().numpy()

    accuracy = (prediction == labels).sum() / len(prediction)
    micro_f1 = f1_score(labels, prediction, average='micro')
    macro_f1 = f1_score(labels, prediction, average='macro')

    return accuracy, micro_f1, macro_f1

@register_trainer("NodeClassificationHet")
class NodeClassificationHetTrainer(BaseNodeClassificationHetTrainer):
    """
    The heterogeneous node classification trainer.

    Parameters
    ----------
    model: ``autogl.module.model.BaseAutoModel``
        Currently Heterogeneous trainer doesn't support decoupled model setting.

    num_features: ``int`` (Optional)
        The number of features in dataset. default None
    
    num_classes: ``int`` (Optional)
        The number of classes. default None

    optimizer: ``Optimizer`` of ``str``
        The (name of) optimizer used to train and predict. default torch.optim.Adam

    lr: ``float``
        The learning rate of node classification task. default 1e-4

    max_epoch: ``int``
        The max number of epochs in training. default 100

    early_stopping_round: ``int``
        The round of early stop. default 100

    weight_decay: ``float``
        weight decay ratio, default 1e-4

    device: ``torch.device`` or ``str``
        The device where model will be running on.

    init: ``bool``
        If True(False), the model will (not) be initialized.

    feval: (Sequence of) ``Evaluation`` or ``str``
        The evaluation functions, default ``[LogLoss]``
    
    loss: ``str``
        The loss used. Default ``nll_loss``.

    lr_scheduler_type: ``str`` (Optional)
        The lr scheduler type used. Default None.

    """

    def __init__(
        self,
        model: Union[BaseAutoModel, str] = None,
        dataset = None,
        num_features=None,
        num_classes=None,
        optimizer=torch.optim.AdamW,
        lr=1e-4,
        max_epoch=100,
        early_stopping_round=100,
        weight_decay=1e-4,
        device="auto",
        init=False,
        feval=[Logloss],
        loss="nll_loss",
        lr_scheduler_type=None,
        *args,
        **kwargs
    ):
        super().__init__(
            model,
            dataset,
            num_features,
            num_classes,
            device=device,
            feval=feval,
            loss=loss,
        )
       
        self.opt_received = optimizer
        if isinstance(optimizer, str):
            if optimizer.lower() == "adam": self.optimizer = torch.optim.Adam
            elif optimizer.lower() == "sgd": self.optimizer = torch.optim.SGD
            else: raise ValueError("Currently not support optimizer {}".format(optimizer))
        elif isinstance(optimizer, type) and issubclass(optimizer, torch.optim.Optimizer):
            self.optimizer = optimizer
        else:
            raise ValueError("Currently not support optimizer {}".format(optimizer))

        self.lr_scheduler_type = lr_scheduler_type

        self.lr = lr
        self.max_epoch = max_epoch
        self.early_stopping_round = early_stopping_round
        self.args = args
        self.kwargs = kwargs
        self.weight_decay = weight_decay

        self.early_stopping = EarlyStopping(
            patience=early_stopping_round, verbose=False
        )

        self.valid_result = None
        self.valid_result_prob = None
        self.valid_score = None

        self.pyg_dgl = DependentBackend.get_backend_name()

        self.hyper_parameter_space = [
            {
                "parameterName": "max_epoch",
                "type": "INTEGER",
                "maxValue": 500,
                "minValue": 10,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "early_stopping_round",
                "type": "INTEGER",
                "maxValue": 30,
                "minValue": 10,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "lr",
                "type": "DOUBLE",
                "maxValue": 1e-1,
                "minValue": 1e-4,
                "scalingType": "LOG",
            },
            {
                "parameterName": "weight_decay",
                "type": "DOUBLE",
                "maxValue": 1e-2,
                "minValue": 1e-4,
                "scalingType": "LOG",
            },
        ]

        self.hyper_parameters = {
            "max_epoch": self.max_epoch,
            "early_stopping_round": self.early_stopping_round,
            "lr": self.lr,
            "weight_decay": self.weight_decay,
        }

        if init is True:
            self.initialize()

    def _initialize(self):
        self.encoder.initialize()

    @classmethod
    def get_task_name(cls):
        """
        Get task name ("NodeClassificationHet")
        """
        return "NodeClassificationHet"

    def _train_only(self, dataset, train_mask="train"):
        G = dataset[0].to(self.device)
        field = dataset.schema["target_node_type"]
        labels = G.nodes[field].data['label'].to(self.device)
        train_mask = self._get_mask(dataset, train_mask).to(self.device)
        val_mask = self._get_mask(dataset, "val").to(self.device)
        model = self.encoder.model.to(self.device)
        optimizer = self.optimizer(
            model.parameters(), lr=self.lr, weight_decay=self.weight_decay
        )

        lr_scheduler_type = self.lr_scheduler_type
        if type(lr_scheduler_type) == str and lr_scheduler_type == "steplr":
            scheduler = StepLR(optimizer, step_size=100, gamma=0.1)
        elif type(lr_scheduler_type) == str and lr_scheduler_type == "multisteplr":
            scheduler = MultiStepLR(optimizer, milestones=[30, 80], gamma=0.1)
        elif type(lr_scheduler_type) == str and lr_scheduler_type == "exponentiallr":
            scheduler = ExponentialLR(optimizer, gamma=0.1)
        elif (
            type(lr_scheduler_type) == str and lr_scheduler_type == "reducelronplateau"
        ):
            scheduler = ReduceLROnPlateau(optimizer, "min")
        else:
            scheduler = None

        for epoch in range(1, self.max_epoch):
            model.train()
            optimizer.zero_grad()
            logits = model(G)

            if hasattr(F, self.loss):
                loss = getattr(F, self.loss)(logits[train_mask], labels[train_mask])
            else:
                raise TypeError(
                    "PyTorch does not support loss type {}".format(self.loss)
                )

            loss.backward()
            optimizer.step()
            if self.lr_scheduler_type:
                scheduler.step()

            if val_mask is not None:
                if type(self.feval) is list:
                    feval = self.feval[0]
                else:
                    feval = self.feval

                val_loss = self.evaluate(dataset, mask=val_mask, feval=feval)
                if feval.is_higher_better() is True:
                    val_loss = -val_loss

                self.early_stopping(val_loss, model)
                if self.early_stopping.early_stop:
                    LOGGER.debug("Early stopping at %d", epoch)
                    break
        if val_mask is not None:
            self.early_stopping.load_checkpoint(model)

    def _predict_only(self, dataset, mask=None):
        model = self.encoder.model.to(self.device)
        model.eval()
        G = dataset[0].to(self.device)
        with torch.no_grad():
            res = model(G)

        if mask is None:
            return res
        else:
            return res[mask]

    def train(self, dataset, keep_valid_result=True, train_mask="train"):
        """
        The function of training on the given dataset and keeping valid result.
        Parameters
        ----------
        dataset: The node classification dataset used to be trained.
        keep_valid_result: ``bool``
            If True(False), save the validation result after training.
        train_mask: The mask for training data
        Returns
        -------
        self: ``autogl.train.NodeClassificationTrainer``
            A reference of current trainer.
        """
        self._train_only(dataset, train_mask)
        G = dataset[0].to(self.device)
        if keep_valid_result:
            # generate labels
            val_mask = G.nodes[dataset.schema["target_node_type"]].data["val_mask"]
            self.valid_result = self._predict_only(dataset)[val_mask].max(1)[1]
            self.valid_result_prob = self._predict_only(dataset)[val_mask]
            self.valid_score = self.evaluate(
                dataset, mask=val_mask, feval=self.feval
            )
            # print(self.valid_score)

    def predict(self, dataset, mask="test"):
        """
        The function of predicting on the given dataset.
        Parameters
        ----------
        dataset: The node classification dataset used to be predicted.
        mask: ``train``, ``val``, or ``test``.
            The dataset mask.
        Returns
        -------
        The prediction result of ``predict_proba``.
        """
        return self.predict_proba(dataset, mask=mask, in_log_format=True).max(1)[1]

    def predict_proba(self, dataset, mask="test", in_log_format=False):
        """
        The function of predicting the probability on the given dataset.
        Parameters
        ----------
        dataset: The node classification dataset used to be predicted.
        mask: ``train``, ``val``, ``test``, or ``Tensor``.
            The dataset mask.
        in_log_format: ``bool``.
            If True(False), the probability will (not) be log format.
        Returns
        -------
        The prediction result.
        """
        G = dataset[0].to(self.device)
        if mask in ["train", "val", "test"]:
            mask = G.nodes[dataset.schema["target_node_type"]].data[f"{mask}_mask"]
        ret = self._predict_only(dataset, mask)
        if in_log_format is True:
            return ret
        else:
            return torch.exp(ret)

    def get_valid_predict(self):
        # """Get the valid result."""
        return self.valid_result

    def get_valid_predict_proba(self):
        # """Get the valid result (prediction probability)."""
        return self.valid_result_prob

    def get_valid_score(self, return_major=True):
        """
        The function of getting the valid score.

        Parameters
        ----------
        return_major: ``bool``.
            If True, the return only consists of the major result.
            If False, the return consists of the all results.

        Returns
        -------
        result: The valid score in training stage.

        """
        if isinstance(self.feval, list):
            if return_major:
                return self.valid_score[0], self.feval[0].is_higher_better()
            else:
                return self.valid_score, [f.is_higher_better() for f in self.feval]
        else:
            return self.valid_score, self.feval.is_higher_better()

    def __repr__(self) -> str:
        import yaml

        return yaml.dump(
            {
                "trainer_name": self.__class__.__name__,
                "optimizer": self.optimizer,
                "learning_rate": self.lr,
                "max_epoch": self.max_epoch,
                "early_stopping_round": self.early_stopping_round,
                "model": repr(self.model.model),
            }
        )
    
    def _get_mask(self, dataset, mask):
        if mask in ["train", "val", "test"]:
            return dataset[0].nodes[dataset.schema["target_node_type"]].data[f"{mask}_mask"]
        return mask

    def evaluate(self, dataset, mask='val', feval = None):
        """
        The function of training on the given dataset and keeping valid result.
        Parameters
        ----------
        dataset: The node classification dataset used to be evaluated.
        mask: ``train``, ``val``, or ``test``.
            The dataset mask.
        feval: ``str``.
            The evaluation method used in this function.
        Returns
        -------
        res: The evaluation result on the given dataset.
        """
        G = dataset[0].to(self.device)

        mask = self._get_mask(dataset, mask)
        label = G.nodes[dataset.schema["target_node_type"]].data['label'].to(self.device)

        if feval is None:
            feval = self.feval
        else:
            feval = get_feval(feval)

        y_pred_prob = self.predict_proba(dataset, mask)

        y_true = label[mask] if mask is not None else label

        if not isinstance(feval, list):
            feval = [feval]
            return_signle = True
        else:
            return_signle = False

        res = []
        for f in feval:
            try:
                res.append(f.evaluate(y_pred_prob, y_true))
            except:
                res.append(f.evaluate(y_pred_prob.cpu().numpy(), y_true.cpu().numpy()))
        if return_signle:
            return res[0]
        return res

    def to(self, new_device):
        self.device = new_device
        if self.model is not None:
            self.model.to(self.device)

    def duplicate_from_hyper_parameter(self, hp: dict, model=None, restricted=True):
        """
        The function of duplicating a new instance from the given hyperparameter.
        
        Parameters
        ----------
        hp: ``dict``.
            The hyperparameter used in the new instance. Should contain 2 keys "trainer", "encoder"
            with corresponding hyperparameters as values.
        model: ``autogl.module.model.BaseAutoModel``
            Currently Heterogeneous trainer doesn't support decoupled model setting.
            If only encoder is specified, decoder will be default to "logsoftmax"

        restricted: ``bool``.
            If False(True), the hyperparameter should (not) be updated from origin hyperparameter.
        
        Returns
        -------
        self: ``autogl.train.NodeClassificationTrainer``
            A new instance of trainer.
        
        """
        trainer_hp = hp["trainer"]
        model_hp = hp["encoder"]
        if not restricted:
            origin_hp = deepcopy(self.hyper_parameters)
            origin_hp.update(trainer_hp)
            trainer_hp = origin_hp
        if model is None:
            model = self.model
        model = model.from_hyper_parameter(model_hp)

        ret = self.__class__(
            model=model,
            dataset=self._dataset,
            num_features=self.num_features,
            num_classes=self.num_classes,
            optimizer=self.opt_received,
            lr=trainer_hp["lr"],
            max_epoch=trainer_hp["max_epoch"],
            early_stopping_round=trainer_hp["early_stopping_round"],
            device=self.device,
            weight_decay=trainer_hp["weight_decay"],
            feval=self.feval,
            loss=self.loss,
            lr_scheduler_type=self.lr_scheduler_type,
            init=True,
            *self.args,
            **self.kwargs
        )

        return ret



import torch
import typing as _typing
import networkx
from networkx.algorithms.euler import is_eulerian
from networkx.algorithms.efficiency_measures import global_efficiency
from networkx.algorithms.efficiency_measures import local_efficiency
from networkx.algorithms.distance_regular import is_distance_regular
from networkx.algorithms.components import number_connected_components
from networkx.algorithms.components import is_connected
# from networkx.algorithms.cluster import average_clustering
from networkx.algorithms.cluster import transitivity
from networkx.algorithms.clique import graph_number_of_cliques
from networkx.algorithms.clique import graph_clique_number
from networkx.algorithms.bridges import has_bridges
from networkx.algorithms.assortativity import degree_pearson_correlation_coefficient
from networkx.algorithms.assortativity import degree_assortativity_coefficient
from networkx.algorithms.approximation.clustering_coefficient import average_clustering
from networkx.algorithms.approximation.clique import large_clique_size

from autogl.data.graph import GeneralStaticGraph
from autogl.data.graph.utils import conversion
from .._base_feature_engineer import BaseFeatureEngineer
from .._feature_engineer_registry import FeatureEngineerUniversalRegistry


class _NetworkXGraphFeatureEngineer(BaseFeatureEngineer):
    def __init__(self, feature_extractor: _typing.Callable[[networkx.Graph], _typing.Any]):
        self.__feature_extractor: _typing.Callable[[networkx.Graph], _typing.Any] = feature_extractor
        super(_NetworkXGraphFeatureEngineer, self).__init__()

    def __transform_homogeneous_static_graph(
            self, homogeneous_static_graph: GeneralStaticGraph
    ) -> GeneralStaticGraph:
        if not (
                homogeneous_static_graph.nodes.is_homogeneous and
                homogeneous_static_graph.edges.is_homogeneous
        ):
            raise ValueError("Provided static graph must be homogeneous")
        dsc: torch.Tensor = torch.tensor(
            [
                self.__feature_extractor(
                    conversion.HomogeneousStaticGraphToNetworkX(to_undirected=True)(homogeneous_static_graph)
                )
            ]
        ).view(-1)
        if 'gf' in homogeneous_static_graph.data:
            gf = homogeneous_static_graph.data['gf'].view(-1)
            homogeneous_static_graph.data['gf'] = torch.cat([gf, dsc])
        else:
            homogeneous_static_graph.data['gf'] = dsc
        return homogeneous_static_graph

    @classmethod
    def __edge_index_to_nx_graph(cls, edge_index: torch.Tensor) -> networkx.Graph:
        g: networkx.Graph = networkx.Graph()
        for u, v in edge_index.t().tolist():
            if u == v:
                continue
            else:
                g.add_edge(u, v)
        return g

    def __transform_data(self, data):
        if not (
                hasattr(data, "edge_index") and
                torch.is_tensor(data.edge_index) and
                isinstance(data.edge_index, torch.Tensor) and
                data.edge_index.dim() == data.edge_index.size(0) == 2 and
                data.edge_index.dtype == torch.long
        ):
            raise TypeError("Unsupported provided data")
        dsc: torch.Tensor = torch.tensor(
            [self.__feature_extractor(self.__edge_index_to_nx_graph(data.edge_index))]
        ).view(-1)
        if hasattr(data, 'gf') and isinstance(data.gf, torch.Tensor):
            gf = data.gf.view(-1)
            data.gf = torch.cat([gf, dsc])
        else:
            data.gf = dsc
        return data

    def _transform(self, data):
        if isinstance(data, GeneralStaticGraph):
            return self.__transform_homogeneous_static_graph(data)
        else:
            return self.__transform_data(data)


@FeatureEngineerUniversalRegistry.register_feature_engineer("NXLargeCliqueSize")
class NXLargeCliqueSize(_NetworkXGraphFeatureEngineer):
    def __init__(self):
        super(NXLargeCliqueSize, self).__init__(large_clique_size)


# @FeatureEngineerUniversalRegistry.register_feature_engineer("NXAverageClusteringApproximate")
# class NXAverageClusteringApproximate(_NetworkXGraphFeatureEngineer):
#     def __init__(self):
#         super(NXAverageClusteringApproximate, self).__init__(average_clustering)


@FeatureEngineerUniversalRegistry.register_feature_engineer("NXDegreeAssortativityCoefficient")
class NXDegreeAssortativityCoefficient(_NetworkXGraphFeatureEngineer):
    def __init__(self):
        super(NXDegreeAssortativityCoefficient, self).__init__(degree_assortativity_coefficient)


@FeatureEngineerUniversalRegistry.register_feature_engineer("NXDegreePearsonCorrelationCoefficient")
class NXDegreePearsonCorrelationCoefficient(_NetworkXGraphFeatureEngineer):
    def __init__(self):
        super(NXDegreePearsonCorrelationCoefficient, self).__init__(degree_pearson_correlation_coefficient)


@FeatureEngineerUniversalRegistry.register_feature_engineer("NXHasBridges")
class NXHasBridges(_NetworkXGraphFeatureEngineer):
    def __init__(self):
        super(NXHasBridges, self).__init__(has_bridges)


@FeatureEngineerUniversalRegistry.register_feature_engineer("NXGraphCliqueNumber")
class NXGraphCliqueNumber(_NetworkXGraphFeatureEngineer):
    def __init__(self):
        super(NXGraphCliqueNumber, self).__init__(graph_clique_number)


@FeatureEngineerUniversalRegistry.register_feature_engineer("NXGraphNumberOfCliques")
class NXGraphNumberOfCliques(_NetworkXGraphFeatureEngineer):
    def __init__(self):
        super(NXGraphNumberOfCliques, self).__init__(graph_number_of_cliques)


@FeatureEngineerUniversalRegistry.register_feature_engineer("NXTransitivity")
class NXTransitivity(_NetworkXGraphFeatureEngineer):
    def __init__(self):
        super(NXTransitivity, self).__init__(transitivity)


@FeatureEngineerUniversalRegistry.register_feature_engineer("NXAverageClustering")
class NXAverageClustering(_NetworkXGraphFeatureEngineer):
    def __init__(self):
        super(NXAverageClustering, self).__init__(average_clustering)


@FeatureEngineerUniversalRegistry.register_feature_engineer("NXIsConnected")
class NXIsConnected(_NetworkXGraphFeatureEngineer):
    def __init__(self):
        super(NXIsConnected, self).__init__(is_connected)


@FeatureEngineerUniversalRegistry.register_feature_engineer("NXNumberConnectedComponents")
class NXNumberConnectedComponents(_NetworkXGraphFeatureEngineer):
    def __init__(self):
        super(NXNumberConnectedComponents, self).__init__(number_connected_components)


@FeatureEngineerUniversalRegistry.register_feature_engineer("NXIsDistanceRegular")
class NXIsDistanceRegular(_NetworkXGraphFeatureEngineer):
    def __init__(self):
        super(NXIsDistanceRegular, self).__init__(is_distance_regular)


@FeatureEngineerUniversalRegistry.register_feature_engineer("NXLocalEfficiency")
class NXLocalEfficiency(_NetworkXGraphFeatureEngineer):
    def __init__(self):
        super(NXLocalEfficiency, self).__init__(local_efficiency)


@FeatureEngineerUniversalRegistry.register_feature_engineer("NXGlobalEfficiency")
class NXGlobalEfficiency(_NetworkXGraphFeatureEngineer):
    def __init__(self):
        super(NXGlobalEfficiency, self).__init__(global_efficiency)


@FeatureEngineerUniversalRegistry.register_feature_engineer("NXIsEulerian")
class NXIsEulerian(_NetworkXGraphFeatureEngineer):
    def __init__(self):
        super(NXIsEulerian, self).__init__(is_eulerian)



import numpy as np
import torch
import typing as _typing
from autogl.data.graph import GeneralStaticGraph
from .._base_feature_engineer import BaseFeatureEngineer
from .._feature_engineer_registry import FeatureEngineerUniversalRegistry


class BaseFeatureSelector(BaseFeatureEngineer):
    def __init__(self):
        self._selection : _typing.Optional[torch.Tensor] = None
        super(BaseFeatureSelector, self).__init__()

    def __transform_homogeneous_static_graph(
            self, static_graph: GeneralStaticGraph
    ) -> GeneralStaticGraph:
        if (
                'x' in static_graph.nodes.data and
                isinstance(self._selection, (torch.Tensor, np.ndarray))
        ):
            static_graph.nodes.data['x'] = static_graph.nodes.data['x'][:, self._selection]
        if (
                'feat' in static_graph.nodes.data and
                isinstance(self._selection, (torch.Tensor, np.ndarray))
        ):
            static_graph.nodes.data['feat'] = static_graph.nodes.data['feat'][:, self._selection]
        return static_graph

    def _transform(
            self, data: _typing.Union[GeneralStaticGraph, _typing.Any]
    ) -> _typing.Union[GeneralStaticGraph, _typing.Any]:
        if isinstance(data, GeneralStaticGraph):
            return self.__transform_homogeneous_static_graph(data)
        elif (
                hasattr(data, 'x') and isinstance(data.x, torch.Tensor) and
                torch.is_tensor(data.x) and data.x.dim() == 2
        ):
            data.x = data.x[:, self._selection]
            return data
        else:
            return data


@FeatureEngineerUniversalRegistry.register_feature_engineer("FilterConstant")
class FilterConstant(BaseFeatureSelector):
    r"""drop constant features"""

    def _fit(self, static_graph: GeneralStaticGraph) -> GeneralStaticGraph:
        if (
                'x' in static_graph.nodes.data and
                self._selection not in (Ellipsis, None) and
                isinstance(self._selection, torch.Tensor) and
                torch.is_tensor(self._selection) and self._selection.dim() == 1
        ):
            feature: _typing.Optional[np.ndarray] = static_graph.nodes.data['x'].numpy()
        elif (
                'feat' in static_graph.nodes.data and
                self._selection not in (Ellipsis, None) and
                isinstance(self._selection, torch.Tensor) and
                torch.is_tensor(self._selection) and self._selection.dim() == 1
        ):
            feature: _typing.Optional[np.ndarray] = static_graph.nodes.data['feat'].numpy()
        else:
            feature: _typing.Optional[np.ndarray] = None
        self._selection: _typing.Optional[torch.Tensor] = torch.from_numpy(
            np.where(np.all(feature == feature[0, :], axis=0) == np.array(False))[0]
            if feature is not None and isinstance(feature, np.ndarray) and feature.ndim == 2
            else None
        )
        return static_graph



import numpy as np
import scipy.sparse as sp
import torch
import torch.utils.data
import typing as _typing
import torch_geometric
from . import target_dependant_sampler


class _LayerDependentImportanceSampler(
    target_dependant_sampler.BasicLayerWiseTargetDependantSampler
):
    """
    Obsolete implementation, unused
    """

    class _Utility:
        @classmethod
        def compute_edge_weights(
            cls, __all_edge_index_with_self_loops: torch.Tensor
        ) -> torch.Tensor:
            __out_degree: torch.Tensor = torch_geometric.utils.degree(
                __all_edge_index_with_self_loops[0]
            )
            __in_degree: torch.Tensor = torch_geometric.utils.degree(
                __all_edge_index_with_self_loops[1]
            )

            temp_tensor: torch.Tensor = torch.stack(
                [
                    __out_degree[__all_edge_index_with_self_loops[0]],
                    __in_degree[__all_edge_index_with_self_loops[1]],
                ]
            )
            temp_tensor: torch.Tensor = torch.pow(temp_tensor, -0.5)
            temp_tensor[torch.isinf(temp_tensor)] = 0.0
            return temp_tensor[0] * temp_tensor[1]

        @classmethod
        def get_candidate_source_nodes_probabilities(
            cls,
            all_candidate_edge_indexes: torch.LongTensor,
            all_edge_index_with_self_loops: torch.Tensor,
            all_edge_weights: torch.Tensor,
        ) -> _typing.Tuple[torch.LongTensor, torch.Tensor]:
            """
            :param all_candidate_edge_indexes:
            :param all_edge_index_with_self_loops: integral edge index with self-loops
            :param all_edge_weights:
            :return: (all_source_nodes_indexes, all_source_nodes_probabilities)
            """
            all_candidate_edge_indexes: torch.LongTensor = (
                all_candidate_edge_indexes.unique()
            )
            _all_candidate_edges_weights: torch.Tensor = all_edge_weights[
                all_candidate_edge_indexes
            ]
            all_candidate_source_nodes_indexes: torch.LongTensor = (
                all_edge_index_with_self_loops[0, all_candidate_edge_indexes].unique()
            )
            all_candidate_source_nodes_probabilities: torch.Tensor = torch.tensor(
                [
                    torch.sum(
                        _all_candidate_edges_weights[
                            all_edge_index_with_self_loops[
                                0, all_candidate_edge_indexes
                            ]
                            == _current_source_node_index
                        ]
                    ).item()
                    / torch.sum(_all_candidate_edges_weights).item()
                    for _current_source_node_index in all_candidate_source_nodes_indexes.tolist()
                ]
            )
            assert (
                all_candidate_source_nodes_indexes.size()
                == all_candidate_source_nodes_probabilities.size()
            )
            return (
                all_candidate_source_nodes_indexes,
                all_candidate_source_nodes_probabilities,
            )

        @classmethod
        def filter_selected_edges_by_source_nodes_and_target_nodes(
            cls,
            all_edges_with_self_loops: torch.Tensor,
            selected_source_node_indexes: torch.LongTensor,
            selected_target_node_indexes: torch.LongTensor,
        ) -> torch.Tensor:
            """
            :param all_edges_with_self_loops: all edges with self loops
            :param selected_source_node_indexes: selected source node indexes
            :param selected_target_node_indexes: selected target node indexes
            :return: filtered edge indexes
            """
            selected_edges_mask_for_source_nodes: torch.Tensor = torch.zeros(
                all_edges_with_self_loops.size(1), dtype=torch.bool
            )
            selected_edges_mask_for_source_nodes[
                torch.cat(
                    [
                        torch.where(
                            all_edges_with_self_loops[0]
                            == __current_selected_source_node_index
                        )[0]
                        for __current_selected_source_node_index in selected_source_node_indexes.unique().tolist()
                    ]
                ).unique()
            ] = True
            selected_edges_mask_for_target_nodes: torch.Tensor = torch.zeros(
                all_edges_with_self_loops.size(1), dtype=torch.bool
            )
            selected_edges_mask_for_target_nodes[
                torch.cat(
                    [
                        torch.where(
                            all_edges_with_self_loops[1]
                            == __current_selected_target_node_index
                        )[0]
                        for __current_selected_target_node_index in selected_target_node_indexes.unique().tolist()
                    ]
                )
            ] = True
            return torch.where(
                selected_edges_mask_for_source_nodes
                & selected_edges_mask_for_target_nodes
            )[0]

    def __init__(
        self,
        edge_index: torch.LongTensor,
        target_nodes_indexes: torch.LongTensor,
        layer_wise_arguments: _typing.Sequence,
        batch_size: _typing.Optional[int] = 1,
        num_workers: int = 0,
        shuffle: bool = True,
        **kwargs
    ):
        super().__init__(
            torch_geometric.utils.add_remaining_self_loops(edge_index)[0],
            target_nodes_indexes,
            layer_wise_arguments,
            batch_size,
            num_workers,
            shuffle,
            **kwargs
        )
        self.__all_edge_weights: torch.Tensor = self._Utility.compute_edge_weights(
            self._edge_index
        )

    def _sample_edges_for_layer(
        self,
        __current_layer_target_nodes_indexes: torch.LongTensor,
        __top_layer_target_nodes_indexes: torch.LongTensor,
        layer_argument: _typing.Any,
        *args,
        **kwargs
    ) -> _typing.Tuple[torch.LongTensor, _typing.Optional[torch.Tensor]]:
        """
        Sample edges for one layer
        :param __current_layer_target_nodes_indexes: target nodes for current layer
        :param __top_layer_target_nodes_indexes: target nodes for top layer
        :param layer_argument: argument for current layer
        :param args: remaining positional arguments
        :param kwargs: remaining keyword arguments
        :return: (edge_id_in_integral_graph, edge_weight)
        """
        if type(layer_argument) != int:
            raise TypeError
        elif not layer_argument > 0:
            raise ValueError
        else:
            sampled_node_size_budget: int = layer_argument

        all_candidate_edge_indexes: torch.LongTensor = torch.cat(
            [
                torch.where(self._edge_index[1] == current_target_node_index)[0]
                for current_target_node_index in __current_layer_target_nodes_indexes.unique().tolist()
            ]
        ).unique()
        (
            __all_candidate_source_nodes_indexes,
            all_candidate_source_nodes_probabilities,
        ) = self._Utility.get_candidate_source_nodes_probabilities(
            all_candidate_edge_indexes,
            self._edge_index,
            self.__all_edge_weights * self.__all_edge_weights,
        )
        assert (
            __all_candidate_source_nodes_indexes.size()
            == all_candidate_source_nodes_probabilities.size()
        )

        """ Sampling """
        if sampled_node_size_budget < __all_candidate_source_nodes_indexes.numel():
            selected_source_node_indexes: torch.LongTensor = (
                __all_candidate_source_nodes_indexes[
                    torch.from_numpy(
                        np.unique(
                            np.random.choice(
                                np.arange(__all_candidate_source_nodes_indexes.numel()),
                                sampled_node_size_budget,
                                p=all_candidate_source_nodes_probabilities.numpy(),
                                replace=False,
                            )
                        )
                    ).unique()
                ].unique()
            )
        else:
            selected_source_node_indexes: torch.LongTensor = (
                __all_candidate_source_nodes_indexes
            )
        selected_source_node_indexes: torch.LongTensor = torch.cat(
            [selected_source_node_indexes, __top_layer_target_nodes_indexes]
        ).unique()

        __selected_edges_indexes: torch.LongTensor = (
            self._Utility.filter_selected_edges_by_source_nodes_and_target_nodes(
                self._edge_index,
                selected_source_node_indexes,
                __current_layer_target_nodes_indexes,
            ).unique()
        )

        non_normalized_selected_edges_weight: torch.Tensor = self.__all_edge_weights[
            __selected_edges_indexes
        ] / torch.tensor(
            [
                all_candidate_source_nodes_probabilities[
                    __all_candidate_source_nodes_indexes == current_source_node_index
                ].item()
                for current_source_node_index in self._edge_index[
                    0, __selected_edges_indexes
                ].tolist()
            ]
        )

        def __normalize_edges_weight_by_target_nodes(
            __edge_index: torch.Tensor, __edge_weight: torch.Tensor
        ) -> torch.Tensor:
            if __edge_index.size(1) != __edge_weight.numel():
                raise ValueError
            for current_target_node_index in __edge_index[1].unique().tolist():
                __current_mask_for_edges: torch.BoolTensor = (
                    __edge_index[1] == current_target_node_index
                )
                __edge_weight[__current_mask_for_edges] = __edge_weight[
                    __current_mask_for_edges
                ] / torch.sum(__edge_weight[__current_mask_for_edges])
            return __edge_weight

        normalized_selected_edges_weight: torch.Tensor = (
            __normalize_edges_weight_by_target_nodes(
                self._edge_index[:, __selected_edges_indexes],
                non_normalized_selected_edges_weight,
            )
        )
        return __selected_edges_indexes, normalized_selected_edges_weight


class LayerDependentImportanceSampler(
    target_dependant_sampler.BasicLayerWiseTargetDependantSampler
):
    """
    The layer-dependent importance sampler from the
    `"Layer-Dependent Importance Sampling for Training Deep and Large Graph Convolutional Networks"
    <https://arxiv.org/abs/1911.07323>`_ literature,  which allows
    for mini-batch training of GNNs on large-scale graphs where full-batch training is not feasible.

    Arguments
    ------------
    edge_index:
        A :obj:`torch.LongTensor` that defines the underlying graph
        connectivity/message passing flow.
        :obj:`edge_index` holds the indices of a (sparse) adjacency matrix.
        If :obj:`edge_index` is of type :obj:`torch.LongTensor`, its shape
        must be defined as :obj:`[2, num_edges]`, where messages from nodes
        :obj:`edge_index[0]` are sent to nodes in :obj:`edge_index[1]`
        (in case :obj:`flow="source_to_target"`).
    target_nodes_indexes:
        indexes of target nodes to learn representation.
    layer_wise_arguments:
        The number of nodes to sample for each layer.
        It's noteworthy that the target nodes for a specific layer
        always be preserved as source nodes for that layer,
        such that the self loops for those target nodes
        are generally preserved for representation learning.
    batch_size:
        number of target nodes for each mini-batch.
    num_workers:
        num_workers argument for inner :class:`torch.utils.data.DataLoader`
    shuffle:
        whether to shuffle target nodes for mini-batches.
    """

    @classmethod
    def __compute_edge_weight(cls, edge_index: torch.Tensor) -> torch.Tensor:
        __num_nodes: int = max(int(edge_index[0].max()), int(edge_index[1].max())) + 1
        _temp_tensor: torch.Tensor = torch.stack(
            [
                torch_geometric.utils.degree(edge_index[0], __num_nodes)[edge_index[0]],
                torch_geometric.utils.degree(edge_index[1], __num_nodes)[edge_index[1]],
            ]
        )
        _temp_tensor: torch.Tensor = torch.pow(_temp_tensor, -0.5)
        _temp_tensor[torch.isinf(_temp_tensor)] = 0
        return _temp_tensor[0] * _temp_tensor[1]

    def __init__(
        self,
        edge_index: torch.LongTensor,
        target_nodes_indexes: torch.LongTensor,
        layer_wise_arguments: _typing.Sequence,
        batch_size: _typing.Optional[int] = 1,
        num_workers: int = 0,
        shuffle: bool = True,
        **kwargs
    ):
        super(LayerDependentImportanceSampler, self).__init__(
            torch_geometric.utils.add_remaining_self_loops(edge_index)[0],
            target_nodes_indexes,
            layer_wise_arguments,
            batch_size,
            num_workers,
            shuffle,
            **kwargs
        )
        self.__edge_weight: torch.Tensor = self.__compute_edge_weight(self._edge_index)
        self.__integral_normalized_l_matrix: sp.csr_matrix = sp.csr_matrix(
            (
                self.__edge_weight.numpy(),
                (self._edge_index[1].numpy(), self._edge_index[0].numpy()),
            )
        )
        self.__integral_edges_indexes_sparse_matrix: sp.csr_matrix = sp.csr_matrix(
            (
                np.arange(self._edge_index.size(1)),
                (self._edge_index[1].numpy(), self._edge_index[0].numpy()),
            )
        )

    def __sample_edges(
        self,
        __current_layer_target_nodes_indexes: np.ndarray,
        __top_layer_target_nodes_indexes: np.ndarray,
        sampled_source_nodes_budget: int,
    ) -> _typing.Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """

        :param __current_layer_target_nodes_indexes: indexes of target nodes for current layer
        :param __top_layer_target_nodes_indexes: indexes of target nodes for top layer
        :param sampled_source_nodes_budget: sampled source nodes budget
        :return: (
                    sampled_edges_indexes,
                    sampled_source_nodes_indexes,
                    corresponding probabilities for sampled_source_nodes_indexes
        )
        """
        partial_l_matrix: sp.csr_matrix = self.__integral_normalized_l_matrix[
            __current_layer_target_nodes_indexes, :
        ]
        p: np.ndarray = np.array(
            np.sum(partial_l_matrix.multiply(partial_l_matrix), axis=0)
        )[0]
        p: np.ndarray = p / np.sum(p)
        _number_of_nodes_to_sample = np.min(
            [np.sum(p > 0), sampled_source_nodes_budget]
        )
        _selected_source_nodes: np.ndarray = np.unique(
            np.concatenate(
                [
                    np.random.choice(
                        p.size, _number_of_nodes_to_sample, replace=False, p=p
                    ),
                    __top_layer_target_nodes_indexes,
                ]
            )
        )

        _sampled_edges_indexes_sparse_matrix: sp.csr_matrix = (
            self.__integral_edges_indexes_sparse_matrix[
                __current_layer_target_nodes_indexes, :
            ]
        )
        _sampled_edges_indexes_sparse_matrix: sp.csc_matrix = (
            _sampled_edges_indexes_sparse_matrix.tocsc()[:, _selected_source_nodes]
        )
        _sampled_edges_indexes: np.ndarray = np.unique(
            _sampled_edges_indexes_sparse_matrix.data
        )

        return _sampled_edges_indexes, _selected_source_nodes, p[_selected_source_nodes]

    def _sample_edges_for_layer(
        self,
        __current_layer_target_nodes_indexes: torch.LongTensor,
        __top_layer_target_nodes_indexes: torch.LongTensor,
        layer_argument: _typing.Any,
        *args,
        **kwargs
    ) -> _typing.Tuple[torch.LongTensor, _typing.Optional[torch.Tensor]]:
        """
        Sample edges for one specific layer, expected to be implemented in subclass.

        Parameters
        ------------
        __current_layer_target_nodes_indexes:
            target nodes for current layer
        __top_layer_target_nodes_indexes:
            target nodes for top layer
        layer_argument:
            argument for current layer
        args:
            remaining positional arguments
        kwargs:
            remaining keyword arguments

        Returns
        --------
        edge_id_in_integral_graph:
            the corresponding positional indexes for the `edge_index` of integral graph
        edge_weight:
            the optional `edge_weight` for aggregation
        """
        __wrapped_result: _typing.Tuple[
            np.ndarray, np.ndarray, np.ndarray
        ] = self.__sample_edges(
            __current_layer_target_nodes_indexes.numpy(),
            __top_layer_target_nodes_indexes.numpy(),
            layer_argument,
        )
        _sampled_edges_indexes: torch.Tensor = torch.from_numpy(__wrapped_result[0])
        _selected_source_nodes: torch.Tensor = torch.from_numpy(__wrapped_result[1])
        _selected_source_nodes_probabilities: torch.Tensor = torch.from_numpy(
            __wrapped_result[2]
        )

        """ Multiply corresponding discount weights """
        __selected_source_node_probability_mapping: _typing.Dict[int, float] = dict(
            zip(
                _selected_source_nodes.tolist(),
                _selected_source_nodes_probabilities.tolist(),
            )
        )
        _selected_edges_weight: torch.Tensor = self.__edge_weight[
            _sampled_edges_indexes
        ]
        _selected_edges_weight: torch.Tensor = _selected_edges_weight / torch.tensor(
            [
                __selected_source_node_probability_mapping.get(
                    _current_source_node_index
                )
                for _current_source_node_index in self._edge_index[
                    0, _sampled_edges_indexes
                ].tolist()
            ]
        )

        """ Normalize edge weight for selected edges by target nodes """
        for _current_target_node_index in (
            self._edge_index[1, _sampled_edges_indexes].unique().tolist()
        ):
            _current_mask_for_selected_edges: torch.BoolTensor = (
                self._edge_index[1, _sampled_edges_indexes]
                == _current_target_node_index
            )
            _selected_edges_weight[
                _current_mask_for_selected_edges
            ] = _selected_edges_weight[_current_mask_for_selected_edges] / torch.sum(
                _selected_edges_weight[_current_mask_for_selected_edges]
            )

        _sampled_edges_indexes: _typing.Union[
            torch.LongTensor, torch.Tensor
        ] = _sampled_edges_indexes
        return _sampled_edges_indexes, _selected_edges_weight



import copy
import torch
import typing
import dgl
import autogl.data.graph
import autogl.data.graph.utils.conversion
from . import _data_preprocessor


class DataPreprocessor(_data_preprocessor.DataPreprocessor):
    @classmethod
    def __preprocess(
            cls, data: typing.Union[dgl.DGLGraph, autogl.data.graph.GeneralStaticGraph, typing.Any]
    ) -> typing.Union[autogl.data.graph.GeneralStaticGraph, typing.Any]:
        if isinstance(data, dgl.DGLGraph):
            graph = autogl.data.graph.utils.conversion.dgl_graph_to_general_static_graph(data)
            setattr(graph, '__ORIGINAL_TYPE_BEFORE_PREPROCESSING', 'DGLGraph')
            return graph
        else:
            return data

    @classmethod
    def __postprocess(
            cls, data: typing.Union[autogl.data.graph.GeneralStaticGraph, typing.Any]
    ) -> typing.Union[dgl.DGLGraph, autogl.data.graph.GeneralStaticGraph, typing.Any]:
        if (
                isinstance(data, autogl.data.graph.GeneralStaticGraph) and
                hasattr(data, '__ORIGINAL_TYPE_BEFORE_PREPROCESSING') and
                isinstance(getattr(data, '__ORIGINAL_TYPE_BEFORE_PREPROCESSING', ...), str) and
                getattr(data, '__ORIGINAL_TYPE_BEFORE_PREPROCESSING', ...) == 'DGLGraph'
        ):
            return autogl.data.graph.utils.conversion.general_static_graph_to_dgl_graph(data)
        else:
            return data

    def fit(self, dataset):
        with torch.no_grad():
            for i, data in enumerate(dataset):
                dataset[i] = self.__postprocess(self._postprocess(self._fit(self._preprocess(self.__preprocess(data)))))
            return dataset

    def transform(self, dataset, inplace: bool = True):
        if not inplace:
            dataset = copy.deepcopy(dataset)
        with torch.no_grad():
            for i, data in enumerate(dataset):
                dataset[i] = self.__postprocess(self._postprocess(self._transform(self._preprocess(self.__preprocess(data)))))
        return dataset

    def fit_transform(self, dataset, inplace: bool = True):
        if not inplace:
            dataset = copy.deepcopy(dataset)
        with torch.no_grad():
            for i, data in enumerate(dataset):
                dataset[i] = self.__postprocess(self._postprocess(self._transform(self._fit(self._preprocess(self.__preprocess(data))))))
        return dataset



import os
from autogl.data.graph import GeneralStaticGraphGenerator
from autogl.data import InMemoryStaticGraphSet
from ._dataset_registry import DatasetUniversalRegistry
import torch_geometric
from torch_geometric.datasets import (
    Amazon, Coauthor, Flickr, ModelNet,
    Planetoid, PPI, QM9, Reddit, TUDataset
)


@DatasetUniversalRegistry.register_dataset("cora")
class CoraDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = Planetoid(os.path.join(path, '_pyg'), "Cora")
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        pyg_data = pyg_dataset[0]

        static_graph = GeneralStaticGraphGenerator.create_homogeneous_static_graph(
            {
                'x': pyg_data.x,
                'y': pyg_data.y,
                'train_mask': getattr(pyg_data, 'train_mask'),
                'val_mask': getattr(pyg_data, 'val_mask'),
                'test_mask': getattr(pyg_data, 'test_mask')
            },
            pyg_data.edge_index
        )
        super(CoraDataset, self).__init__([static_graph])


@DatasetUniversalRegistry.register_dataset("CiteSeer".lower())
class CiteSeerDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = Planetoid(os.path.join(path, '_pyg'), "CiteSeer")
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        pyg_data = pyg_dataset[0]

        static_graph = GeneralStaticGraphGenerator.create_homogeneous_static_graph(
            {
                'x': pyg_data.x,
                'y': pyg_data.y,
                'train_mask': getattr(pyg_data, 'train_mask'),
                'val_mask': getattr(pyg_data, 'val_mask'),
                'test_mask': getattr(pyg_data, 'test_mask')
            },
            pyg_data.edge_index
        )
        super(CiteSeerDataset, self).__init__([static_graph])


@DatasetUniversalRegistry.register_dataset("PubMed".lower())
class PubMedDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = Planetoid(os.path.join(path, '_pyg'), "PubMed")
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        pyg_data = pyg_dataset[0]

        static_graph = GeneralStaticGraphGenerator.create_homogeneous_static_graph(
            {
                'x': pyg_data.x,
                'y': pyg_data.y,
                'train_mask': getattr(pyg_data, 'train_mask'),
                'val_mask': getattr(pyg_data, 'val_mask'),
                'test_mask': getattr(pyg_data, 'test_mask')
            },
            pyg_data.edge_index
        )
        super(PubMedDataset, self).__init__([static_graph])


@DatasetUniversalRegistry.register_dataset("flickr")
class FlickrDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = Flickr(os.path.join(path, '_pyg'))
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        pyg_data = pyg_dataset[0]

        static_graph = GeneralStaticGraphGenerator.create_homogeneous_static_graph(
            {
                'x': pyg_data.x,
                'y': pyg_data.y,
                'train_mask': getattr(pyg_data, 'train_mask'),
                'val_mask': getattr(pyg_data, 'val_mask'),
                'test_mask': getattr(pyg_data, 'test_mask')
            },
            pyg_data.edge_index
        )
        super(FlickrDataset, self).__init__([static_graph])


@DatasetUniversalRegistry.register_dataset("reddit")
class RedditDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = Reddit(os.path.join(path, '_pyg'))
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        pyg_data = pyg_dataset[0]

        static_graph = GeneralStaticGraphGenerator.create_homogeneous_static_graph(
            {
                'x': pyg_data.x,
                'y': pyg_data.y,
                'train_mask': getattr(pyg_data, 'train_mask'),
                'val_mask': getattr(pyg_data, 'val_mask'),
                'test_mask': getattr(pyg_data, 'test_mask')
            },
            pyg_data.edge_index
        )
        super(RedditDataset, self).__init__([static_graph])


@DatasetUniversalRegistry.register_dataset("amazon_computers")
class AmazonComputersDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = Amazon(os.path.join(path, '_pyg'), "Computers")
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        pyg_data = pyg_dataset[0]
        static_graph = GeneralStaticGraphGenerator.create_homogeneous_static_graph(
            {'x': pyg_data.x, 'y': pyg_data.y},
            pyg_data.edge_index
        )
        super(AmazonComputersDataset, self).__init__([static_graph])


@DatasetUniversalRegistry.register_dataset("amazon_photo")
class AmazonPhotoDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = Amazon(os.path.join(path, '_pyg'), "Photo")
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        pyg_data = pyg_dataset[0]
        static_graph = GeneralStaticGraphGenerator.create_homogeneous_static_graph(
            {'x': pyg_data.x, 'y': pyg_data.y},
            pyg_data.edge_index
        )
        super(AmazonPhotoDataset, self).__init__([static_graph])


@DatasetUniversalRegistry.register_dataset("coauthor_physics")
class CoauthorPhysicsDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = Coauthor(os.path.join(path, '_pyg'), "Physics")
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        pyg_data = pyg_dataset[0]
        static_graph = GeneralStaticGraphGenerator.create_homogeneous_static_graph(
            {'x': pyg_data.x, 'y': pyg_data.y},
            pyg_data.edge_index
        )
        super(CoauthorPhysicsDataset, self).__init__([static_graph])


@DatasetUniversalRegistry.register_dataset("coauthor_cs")
class CoauthorCSDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = Coauthor(os.path.join(path, '_pyg'), "CS")
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        pyg_data = pyg_dataset[0]
        static_graph = GeneralStaticGraphGenerator.create_homogeneous_static_graph(
            {'x': pyg_data.x, 'y': pyg_data.y},
            pyg_data.edge_index
        )
        super(CoauthorCSDataset, self).__init__([static_graph])


@DatasetUniversalRegistry.register_dataset("ppi")
class PPIDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        train_dataset = PPI(os.path.join(path, '_pyg'), 'train')
        if hasattr(train_dataset, "__data_list__"):
            delattr(train_dataset, "__data_list__")
        if hasattr(train_dataset, "_data_list"):
            delattr(train_dataset, "_data_list")
        val_dataset = PPI(os.path.join(path, '_pyg'), 'val')
        if hasattr(val_dataset, "__data_list__"):
            delattr(val_dataset, "__data_list__")
        if hasattr(val_dataset, "_data_list"):
            delattr(val_dataset, "_data_list")
        test_dataset = PPI(os.path.join(path, '_pyg'), 'test')
        if hasattr(test_dataset, "__data_list__"):
            delattr(test_dataset, "__data_list__")
        if hasattr(test_dataset, "_data_list"):
            delattr(test_dataset, "_data_list")
        train_index = range(len(train_dataset))
        val_index = range(len(train_dataset), len(train_dataset) + len(val_dataset))
        test_index = range(
            len(train_dataset) + len(val_dataset),
            len(train_dataset) + len(val_dataset) + len(test_dataset)
        )
        super(PPIDataset, self).__init__(
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    {'x': data.x, 'y': data.y}, data.edge_index
                ) for data in train_dataset
            ] +
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    {'x': data.x, 'y': data.y}, data.edge_index
                ) for data in val_dataset
            ] +
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    {'x': data.x, 'y': data.y}, data.edge_index
                ) for data in test_dataset
            ],
            train_index, val_index, test_index
        )


@DatasetUniversalRegistry.register_dataset("qm9")
class QM9Dataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = QM9(os.path.join(path, '_pyg'))
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        super(QM9Dataset, self).__init__(
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    {'x': data.x, 'pos': data.pos, 'z': data.z},
                    data.edge_index,
                    edges_data={'edge_attr': data.edge_attr},
                    graph_data={'idx': data.idx, 'y': data.y}
                ) for data in pyg_dataset
            ]
        )


@DatasetUniversalRegistry.register_dataset("mutag")
class MUTAGDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = TUDataset(os.path.join(path, '_pyg'), "MUTAG")
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        super(MUTAGDataset, self).__init__(
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    {'x': pyg_data.x}, pyg_data.edge_index,
                    edges_data={'edge_attr': pyg_data.edge_attr},
                    graph_data={'y': pyg_data.y}
                )
                for pyg_data in pyg_dataset
            ]
        )


@DatasetUniversalRegistry.register_dataset("enzymes")
class ENZYMESDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = TUDataset(os.path.join(path, '_pyg'), "ENZYMES")
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        super(ENZYMESDataset, self).__init__(
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    {'x': pyg_data.x}, pyg_data.edge_index,
                    graph_data={'y': pyg_data.y}
                )
                for pyg_data in pyg_dataset
            ]
        )


@DatasetUniversalRegistry.register_dataset("imdb-b")
@DatasetUniversalRegistry.register_dataset("imdb-binary")
@DatasetUniversalRegistry.register_dataset("IMDBb".upper())
@DatasetUniversalRegistry.register_dataset("IMDBBinary".upper())
@DatasetUniversalRegistry.register_dataset("IMDBBinary")
class IMDBBinaryDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = TUDataset(os.path.join(path, '_pyg'), "IMDB-BINARY")
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        super(IMDBBinaryDataset, self).__init__(
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    {}, pyg_data.edge_index, graph_data={'y': pyg_data.y}
                )
                for pyg_data in pyg_dataset
            ]
        )


@DatasetUniversalRegistry.register_dataset("imdb-m")
@DatasetUniversalRegistry.register_dataset("imdb-multi")
@DatasetUniversalRegistry.register_dataset("IMDBm".upper())
@DatasetUniversalRegistry.register_dataset("IMDBMulti".upper())
@DatasetUniversalRegistry.register_dataset("IMDBMulti")
class IMDBMultiDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = TUDataset(os.path.join(path, '_pyg'), "IMDB-MULTI")
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        super(IMDBMultiDataset, self).__init__(
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    {}, pyg_data.edge_index, graph_data={'y': pyg_data.y}
                )
                for pyg_data in pyg_dataset
            ]
        )


@DatasetUniversalRegistry.register_dataset("reddit-b")
@DatasetUniversalRegistry.register_dataset("reddit-binary")
@DatasetUniversalRegistry.register_dataset("RedditB".upper())
@DatasetUniversalRegistry.register_dataset("RedditBinary".upper())
class RedditBinaryDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = TUDataset(os.path.join(path, '_pyg'), "REDDIT-BINARY")
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        super(RedditBinaryDataset, self).__init__(
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    {}, pyg_data.edge_index, graph_data={'y': pyg_data.y}
                )
                for pyg_data in pyg_dataset
            ]
        )


@DatasetUniversalRegistry.register_dataset("reddit-multi-5k")
@DatasetUniversalRegistry.register_dataset("RedditMulti5K".upper())
class REDDITMulti5KDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = TUDataset(os.path.join(path, '_pyg'), "REDDIT-MULTI-5K")
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        super(REDDITMulti5KDataset, self).__init__(
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    {}, pyg_data.edge_index, graph_data={'y': pyg_data.y}
                )
                for pyg_data in pyg_dataset
            ]
        )


@DatasetUniversalRegistry.register_dataset("reddit-multi-12k")
@DatasetUniversalRegistry.register_dataset("RedditMulti12K".upper())
class REDDITMulti12KDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = TUDataset(os.path.join(path, '_pyg'), "REDDIT-MULTI-12K")
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        super(REDDITMulti12KDataset, self).__init__(
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    {}, pyg_data.edge_index, graph_data={'y': pyg_data.y}
                )
                for pyg_data in pyg_dataset
            ]
        )


@DatasetUniversalRegistry.register_dataset("collab")
class COLLABDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = TUDataset(os.path.join(path, '_pyg'), "COLLAB")
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        super(COLLABDataset, self).__init__(
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    {}, pyg_data.edge_index, graph_data={'y': pyg_data.y}
                )
                for pyg_data in pyg_dataset
            ]
        )


@DatasetUniversalRegistry.register_dataset("proteins")
class ProteinsDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = TUDataset(os.path.join(path, '_pyg'), "PROTEINS")
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        super(ProteinsDataset, self).__init__(
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    {'x': pyg_data.x}, pyg_data.edge_index, graph_data={'y': pyg_data.y}
                )
                for pyg_data in pyg_dataset
            ]
        )


@DatasetUniversalRegistry.register_dataset("ptc-mr")
class PTCMRDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = TUDataset(os.path.join(path, '_pyg'), "PTC_MR")
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        super(PTCMRDataset, self).__init__(
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    {'x': pyg_data.x}, pyg_data.edge_index,
                    edges_data={'edge_attr': pyg_data.edge_attr},
                    graph_data={'y': pyg_data.y}
                )
                for pyg_data in pyg_dataset
            ]
        )


@DatasetUniversalRegistry.register_dataset("nci1")
class NCI1Dataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = TUDataset(os.path.join(path, '_pyg'), "NCI1")
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        super(NCI1Dataset, self).__init__(
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    {'x': pyg_data.x}, pyg_data.edge_index,
                    graph_data={'y': pyg_data.y}
                )
                for pyg_data in pyg_dataset
            ]
        )


@DatasetUniversalRegistry.register_dataset("nci109")
class NCI109Dataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = TUDataset(os.path.join(path, '_pyg'), "NCI109")
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        super(NCI109Dataset, self).__init__(
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    {'x': pyg_data.x}, pyg_data.edge_index,
                    graph_data={'y': pyg_data.y}
                )
                for pyg_data in pyg_dataset
            ]
        )


@DatasetUniversalRegistry.register_dataset("ModelNet10Training")
class ModelNet10TrainingDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = ModelNet(
            os.path.join(path, '_pyg'), '10', True,
            pre_transform=torch_geometric.transforms.FaceToEdge()
        )
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        super(ModelNet10TrainingDataset, self).__init__(
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    {'pos': pyg_data.pos},
                    pyg_data.edge_index,
                    graph_data={'y': pyg_data.y}
                )
                for pyg_data in pyg_dataset
            ]
        )


@DatasetUniversalRegistry.register_dataset("ModelNet10Test")
class ModelNet10TestDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = ModelNet(
            os.path.join(path, '_pyg'), '10', False,
            pre_transform=torch_geometric.transforms.FaceToEdge()
        )
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        super(ModelNet10TestDataset, self).__init__(
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    {'pos': pyg_data.pos},
                    pyg_data.edge_index,
                    graph_data={'y': pyg_data.y}
                )
                for pyg_data in pyg_dataset
            ]
        )


@DatasetUniversalRegistry.register_dataset("ModelNet40Training")
class ModelNet40TrainingDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = ModelNet(
            os.path.join(path, '_pyg'), '40', True,
            pre_transform=torch_geometric.transforms.FaceToEdge()
        )
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        super(ModelNet40TrainingDataset, self).__init__(
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    {'pos': pyg_data.pos},
                    pyg_data.edge_index,
                    graph_data={'y': pyg_data.y}
                )
                for pyg_data in pyg_dataset
            ]
        )


@DatasetUniversalRegistry.register_dataset("ModelNet40Test")
class ModelNet40TestDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        pyg_dataset = ModelNet(
            os.path.join(path, '_pyg'), '40', False,
            pre_transform=torch_geometric.transforms.FaceToEdge()
        )
        if hasattr(pyg_dataset, "__data_list__"):
            delattr(pyg_dataset, "__data_list__")
        if hasattr(pyg_dataset, "_data_list"):
            delattr(pyg_dataset, "_data_list")
        super(ModelNet40TestDataset, self).__init__(
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    {'pos': pyg_data.pos},
                    pyg_data.edge_index,
                    graph_data={'y': pyg_data.y}
                )
                for pyg_data in pyg_dataset
            ]
        )



from . import (
    backend,
    data,
    datasets,
    module,
    solver,
    utils,
)

from .module import (
    ensemble,
    feature,
    hpo,
    model,
    nas,
    train,
)

__version__ = "0.3.1"



from .operations import *
OPS = {
    'ZERO': lambda indim, outdim, head, dropout, concat=False: Zero(indim, outdim),
    'CONST': lambda indim, outdim, head, dropout, concat=False: GeoLayer(indim, outdim, head, concat, att_type='const', dropout=dropout),
    'GCN': lambda indim, outdim, head, dropout, concat=False: GeoLayer(indim, outdim, head, concat, att_type='gcn', dropout=dropout),
    'GAT': lambda indim, outdim, head, dropout, concat=False: GeoLayer(indim, outdim, head, concat, att_type='gat', dropout=dropout),
    'SYM': lambda indim, outdim, head, dropout, concat=False: GeoLayer(indim, outdim, head, concat, att_type='gat_sym', dropout=dropout),
    'COS': lambda indim, outdim, head, dropout, concat=False: GeoLayer(indim, outdim, head, concat, att_type='cos', dropout=dropout),
    'LIN': lambda indim, outdim, head, dropout, concat=False: GeoLayer(indim, outdim, head, concat, att_type='linear', dropout=dropout),
    'GENE': lambda indim, outdim, head, dropout, concat=False: GeoLayer(indim, outdim, head, concat, att_type='generalized_linear', dropout=dropout)
}

PRIMITIVES = list(OPS.keys())



from . import _abstract_views


class GeneralStaticGraph:
    @property
    def nodes(self) -> _abstract_views.HeterogeneousNodeView:
        raise NotImplementedError

    @property
    def edges(self) -> _abstract_views.HeterogeneousEdgesView:
        raise NotImplementedError

    @property
    def data(self) -> _abstract_views.GraphDataView:
        raise NotImplementedError
